#!/usr/bin/env python
#
# selection.py - The Selection class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`Selection` class, which represents a
selection of voxels in a 3D :class:`.Image`.
"""

import logging
import collections.abc as abc

import numpy                       as np
import scipy.ndimage.measurements  as ndimeas

import fsl.utils.notifier          as notifier
import fsleyes.gl.routines         as glroutines


log = logging.getLogger(__name__)


class Selection(notifier.Notifier):
    """The ``Selection`` class represents a selection of voxels in a 3D
    :class:`.Image`. The selection is stored as a ``numpy`` mask array,
    the same shape as the image. Methods are available to query and update
    the selection.


    Changes to a ``Selection`` can be made through *blocks*, which are 3D
    cuboid regions. The following methods allow a block to be
    selected/deselected, where the block is specified by a voxel coordinate,
    and a block size:

    .. autosummary::
       :nosignatures:

       selectBlock
       deselectBlock


    The following methods offer more fine grained control over selection
    blocks - with these methods, you pass in a block that you have created
    yourself, and an offset into the selection, specifying its location:

    .. autosummary::
       :nosignatures:

       setSelection
       replaceSelection
       addToSelection
       removeFromSelection


    A third approach to making a selection is provided by the
    :meth:`selectByValue` method, which allows a selection to be made
    in a manner similar to a *bucket fill* technique found in any image
    editor.


    The related :meth:`invertRegion` method, given a seed location, will
    invert the selected state of all voxels adjacent to that location. This
    approach allows a *fill holes* type approach, where a region outline is
    delineated, and then the interior inverted to select it.


    A ``Selection`` object keeps track of the most recent change made through
    any of the above methods. The most recent change can be retrieved through
    the :meth:`getLastChange` method. The ``Selection`` class inherits from
    the :class:`.Notifier` class - you can be notified whenever the selection
    changes by registering as a listener.


    Finally, the ``Selection`` class offers a few other methods for
    convenience:

    .. autosummary::
       :nosignatures:

       getSelection
       getSelectionSize
       clearSelection
       getBoundedSelection
       getIndices
    """


    def __init__(self, image, display, selection=None):
        """Create a ``Selection`` instance.

        :arg image:     The :class:`.Image` instance  associated with this
                        ``Selection``.

        :arg display:   The :class:`.Display` instance for the ``image``.

        :arg selection: Selection array. If not provided, one is created.
                        Must be a ``numpy.uint8`` array with the same shape
                        as ``image``. This array is *not* copied.
        """

        self.__image              = image
        self.__display            = display
        self.__opts               = display.opts
        self.__clear              = True
        self.__lastChangeOffset   = None
        self.__lastChangeOldBlock = None
        self.__lastChangeNewBlock = None

        if selection is None:
            selection = np.zeros(image.shape[:3], dtype=np.uint8)

        elif selection.shape != image.shape[:3] or \
             selection.dtype != np.uint8:
            raise ValueError('Incompatible selection array: {} ({})'.format(
                selection.shape, selection.dtype))

        self.__selection = selection

        log.debug('{}.init ({})'.format(type(self).__name__, id(self)))


    def __del__(self):
        """Prints a log message."""
        if log:
            log.debug('{}.del ({})'.format(type(self).__name__, id(self)))


    @property
    def shape(self):
        """Returns the selection shape. """
        return self.__selection.shape


    def getSelection(self):
        """Returns the selection array.

        .. warning:: Do not modify the selection array directly - use the
                     ``Selection`` instance methods
                     (e.g. :meth:`setSelection`) instead.  If you modify the
                     selection directly through this attribute, the
                     :meth:`getLastChange` method, and selection notification,
                     will break.
        """
        return self.__selection


    def selectBlock(self,
                    voxel,
                    boxSize,
                    axes=(0, 1, 2),
                    bias=None,
                    combine=False):
        """Selects the block (sets all voxels to 1) specified by the given
        voxel and block size. See the :func:`.routines.voxelBlock` function
        for details on the arguments.

        :arg combine:   Combine this change with the previous stored change
                        (see :meth:`__storeChange`).
        """

        block, offset = glroutines.voxelBlock(
            voxel,
            self.__selection.shape,
            boxSize,
            bias=bias,
            axes=axes)

        self.addToSelection(block, offset, combine)


    def deselectBlock(self,
                      voxel,
                      boxSize,
                      axes=(0, 1, 2),
                      bias=None,
                      combine=False):
        """De-selects the block (sets all voxels to 0) specified by the given
        voxel and box size. See the :func:`.routines.voxelBlock` function
        for details on the arguments.

        :arg combine:   Combine this change with the previous stored change
                        (see :meth:`__storeChange`).
        """

        block, offset = glroutines.voxelBlock(
            voxel,
            self.__selection.shape,
            boxSize,
            bias=bias,
            axes=axes)

        self.removeFromSelection(block, offset, combine)


    def setSelection(self, block, offset, combine=False):
        """Copies the given ``block`` into the selection, starting at
        ``offset``.

        :arg block:   A ``numpy.uint8`` array containing a selection.

        :arg offset:  Voxel coordinates specifying the block location.

        :arg combine: Combine this change with the previous stored change (see
                      :meth:`__storeChange`).
        """
        self.__updateSelectionBlock(block, offset, combine)


    def replaceSelection(self, block, offset, combine=False):
        """Copies the given ``block`` into the selection, starting at
        ``offset``.

        :arg block:   A ``numpy.uint8`` array containing a selection.

        :arg offset:  Voxel coordinates specifying the block location.

        :arg combine: Combine this change with the previous stored change (see
                      :meth:`__storeChange`).
        """

        self.__updateSelectionBlock(block, offset, combine)


    def addToSelection(self, block, offset, combine=False):
        """Adds the selection (via a boolean OR operation) in the given
        ``block`` to the current selection, starting at ``offset``.

        :arg block:   A ``numpy.uint8`` array containing a selection.

        :arg offset:  Voxel coordinates specifying the block location.

        :arg combine: Combine this change with the previous stored change (see
                      :meth:`__storeChange`).
        """
        existing = self.__getSelectionBlock(block.shape, offset)
        block    = np.logical_or(block, existing)

        self.__updateSelectionBlock(block, offset, combine)


    def removeFromSelection(self, block, offset, combine=False):
        """Clears all voxels in the selection where the values in ``block``
        are non-zero.

        :arg block:   A ``numpy.uint8`` array containing a selection.

        :arg offset:  Voxel coordinates specifying the block location.

        :arg combine: Combine this change with the previous stored change (see
                      :meth:`__storeChange`).
        """
        existing             = self.__getSelectionBlock(block.shape, offset)
        existing[block != 0] = False
        self.__updateSelectionBlock(existing, offset, combine)


    def getSelectionSize(self):
        """Returns the number of voxels that are currently selected. """
        return self.__selection.sum()


    def getBoundedSelection(self):
        """Extracts the smallest region from the :attr:`selection` which
        contains all selected voxels.

        Returns a tuple containing the region, as a ``numpy.uint8`` array, and
        the coordinates specifying its location in the full :attr:`selection`
        array.

        .. warning:: This method is slow, and in many cases it may be
                     faster simply to access the full selection array.
        """

        xs, ys, zs = np.where(self.__selection > 0)

        if len(xs) == 0:
            return np.array([]).reshape((0, 0, 0)), (0, 0, 0)

        xlo = int(xs.min())
        ylo = int(ys.min())
        zlo = int(zs.min())
        xhi = int(xs.max() + 1)
        yhi = int(ys.max() + 1)
        zhi = int(zs.max() + 1)

        selection = self.__selection[xlo:xhi, ylo:yhi, zlo:zhi]

        return selection, (xlo, ylo, zlo)


    def clearSelection(self, restrict=None, combine=False):
        """Clears (sets to 0) the entire selection, or the selection specified
        by the ``restrict`` parameter, if it is given.

        .. note:: Calling this method when the selection is already empty
                  will clear the most recently stored change - see
                  :meth:`getLastChange`.

        :arg restrict: An optional sequence of three ``slice`` objects,
                       specifying the portion of the selection to clear.

        :arg combine:  Combine this change with the previous stored change (see
                       :meth:`__storeChange`).
        """

        if self.__clear:
            self.setChange(None, None)
            return

        fRestrict = fixSlices(restrict)
        offset    = [r.start if r.start is not None else 0 for r in fRestrict]

        log.debug('Clearing selection ({}): {}'.format(id(self), fRestrict))

        block                       = np.array(self.__selection[fRestrict])
        self.__selection[fRestrict] = False

        self.__storeChange(block,
                           np.array(self.__selection[fRestrict]),
                           offset,
                           combine)

        # Set the internal clear flag to True,
        # when the entire selection has been
        # cleared, so we can skip subsequent
        # redundant clears.
        if restrict is None:
            self.__clear = True

        self.notify()


    def getLastChange(self):
        """Returns the most recent change made to this ``Selection``.

        A tuple is returned, containing the following:

         - A ``numpy.uint8`` array containing the old block value
         - A ``numpy.uint8`` array containing the new block value
         - Voxel coordinates denoting the block location in the full
           :attr:`selection` array.

        If there is no stored change this method will return ``(None, None,
        None)`` (see also the note in :meth:`clearSelection`).
        """

        return (self.__lastChangeOldBlock,
                self.__lastChangeNewBlock,
                self.__lastChangeOffset)


    def setChange(self, block, offset, oldBlock=None):
        """Sets/overwrites the most recently saved change made to this
        ``Selection``.
        """
        self.__lastChangeOldBlock = oldBlock
        self.__lastChangeNewBlock = block
        self.__lastChangeOffset   = offset


    def __storeChange(self, old, new, offset, combine=False):
        """Stores the given selection change.

        :arg old:     A copy of the portion of the :attr:`selection` that
                      has changed,

        :arg new:     The new selection values.

        :arg offset:  Offset into the full :attr:`selection` array

        :arg combine: If ``False`` (the default), the previously stored change
                      will be replaced by the current change. Otherwise the
                      previous and current changes will be combined.
        """

        # Not combining changes (or there
        # is no previously stored change).
        # We store the change, replacing
        # the previous one.
        if (not combine) or (self.__lastChangeNewBlock is None):

            if log.getEffectiveLevel() == logging.DEBUG:
                log.debug('Replacing previously stored change with: '
                          '[({}, {}), ({}, {}), ({}, {})] ({} selected)'
                          .format(offset[0], offset[0] + old.shape[0],
                                  offset[1], offset[1] + old.shape[1],
                                  offset[2], offset[2] + old.shape[2],
                                  new.sum()))

            self.__lastChangeOldBlock = old
            self.__lastChangeNewBlock = new
            self.__lastChangeOffset   = offset
            return

        # Otherwise, we combine the old
        # change with the new one.
        lcOld     = self.__lastChangeOldBlock
        lcNew     = self.__lastChangeNewBlock
        lcOffset  = self.__lastChangeOffset

        # The old block might be None, which
        # implies all zeros
        if lcOld is None:
            lcOld = np.zeros(lcNew.shape, dtype=lcNew.dtype)

        # Calculate/organise low/high indices
        # for each change set:
        #
        #  - one for the current change (passed
        #    in to this method call)
        #
        #  - One for the last stored change
        #
        #  - One for the combination of the above
        currIdxs = []
        lastIdxs = []
        cmbIdxs  = []

        for ax in range(3):

            currLo = offset[  ax]
            lastLo = lcOffset[ax]
            currHi = offset[  ax] + old  .shape[ax]
            lastHi = lcOffset[ax] + lcOld.shape[ax]

            cmbLo  = min(currLo, lastLo)
            cmbHi  = max(currHi, lastHi)

            currIdxs.append((int(currLo), int(currHi)))
            lastIdxs.append((int(lastLo), int(lastHi)))
            cmbIdxs .append((int(cmbLo),  int(cmbHi)))

        # Make slice objects for each of the indices,
        # to make indexing easier. The last/current
        # slice objects are defined relative to the
        # combined space of both.
        cmbSlices  = tuple([slice(lo, hi) for lo, hi in cmbIdxs])

        lastSlices = tuple([slice(lLo - cmLo, lHi - cmLo)
                            for ((lLo, lHi), (cmLo, cmHi))
                            in zip(lastIdxs, cmbIdxs)])

        currSlices = tuple([slice(cuLo - cmLo, cuHi - cmLo)
                            for ((cuLo, cuHi), (cmLo, cmHi))
                            in zip(currIdxs, cmbIdxs)])

        cmbOld    = np.array(self.__selection[cmbSlices])
        cmbNew    = np.array(cmbOld)

        cmbOld[lastSlices] = lcOld
        cmbNew[lastSlices] = lcNew
        cmbNew[currSlices] = new

        if log.getEffectiveLevel() == logging.DEBUG:
            log.debug('Combining changes: '
                      '[({}, {}), ({}, {}), ({}, {})] ({} selected) + '
                      '[({}, {}), ({}, {}), ({}, {})] ({} selected) = '
                      '[({}, {}), ({}, {}), ({}, {})] ({} selected)'.format(
                          lastIdxs[0][0], lastIdxs[0][1],
                          lastIdxs[1][0], lastIdxs[1][1],
                          lastIdxs[2][0], lastIdxs[2][1],
                          lcNew.sum(),
                          currIdxs[0][0], currIdxs[0][1],
                          currIdxs[1][0], currIdxs[1][1],
                          currIdxs[2][0], currIdxs[2][1],
                          new.sum(),
                          cmbIdxs[0][0], cmbIdxs[0][1],
                          cmbIdxs[1][0], cmbIdxs[1][1],
                          cmbIdxs[2][0], cmbIdxs[2][1],
                          cmbNew.sum()))

        self.__lastChangeOldBlock = cmbOld
        self.__lastChangeNewBlock = cmbNew
        self.__lastChangeOffset   = cmbIdxs[0][0], cmbIdxs[1][0], cmbIdxs[2][0]


    def getIndices(self, restrict=None):
        """Returns a :math:`N \\times 3` array which contains the
        coordinates of all voxels that are currently selected.

        If the ``restrict`` argument is not provided, the entire
        selection image is searched.

        :arg restrict: A ``slice`` object specifying a sub-set of the
                       full selection to consider.
        """

        restrict   = fixSlices(restrict)
        xs, ys, zs = np.where(self.__selection[restrict])
        result     = np.vstack((xs, ys, zs)).T

        for ax in range(3):

            off = restrict[ax].start

            if off is not None:
                result[:, ax] += off

        return result


    def selectByValue(self,
                      seedLoc,
                      precision=None,
                      searchRadius=None,
                      local=False,
                      restrict=None,
                      combine=False):
        """A *bucket fill* style selection routine.

        :arg combine:      Combine with the previous stored change (see
                           :meth:`__storeChange`).

        See the :func:`selectByValue` function for details on the other
        arguments.

        :returns: The generated selection array (a ``numpy`` boolean array),
                  and offset of this array into the full selection image.
        """

        data = self.__image[self.__opts.index()]
        block, offset = selectByValue(data,
                                      seedLoc,
                                      precision,
                                      searchRadius,
                                      local,
                                      restrict)

        self.replaceSelection(block, offset, combine)

        return block, offset


    def invertRegion(self, seedLoc, restrict=None):
        """Inverts the selected state of the region adjacent to ``seedLoc``.

        See the :func:`selectByValue` function for details on the other
        arguments.
        """

        data = self.__selection
        val  = data[tuple(seedLoc)]
        block, offset = selectByValue(data,
                                      seedLoc,
                                      0.5,
                                      local=True,
                                      restrict=restrict)

        if val == 0: self.addToSelection(     block, offset)
        else:        self.removeFromSelection(block, offset)

        return block, offset


    def selectLine(self,
                   from_,
                   to,
                   boxSize,
                   axes=(0, 1, 2),
                   bias=None,
                   combine=False):
        """Selects a line from ``from_`` to ``to``.

        :arg combine: Combine with the previous stored change (see
                      :meth:`__storeChange`).

        See the :func:`selectLine` function for details on the other arguments.
        """

        block, offset = selectLine(self.__selection.shape,
                                   self.__image.pixdim[:3],
                                   from_,
                                   to,
                                   boxSize,
                                   axes,
                                   bias)

        self.addToSelection(block, offset, combine)

        return block, offset


    def deselectLine(self,
                     from_,
                     to,
                     boxSize,
                     axes=(0, 1, 2),
                     bias=None,
                     combine=False):
        """Deselects a line from ``from_`` to ``to``.

        :arg combine: Combine with the previous stored change (see
                      :meth:`__storeChange`).

        See the :func:`selectLine` function for details on the other arguments.
        """

        block, offset = selectLine(self.__selection.shape,
                                   self.__image.pixdim[:3],
                                   from_,
                                   to,
                                   boxSize,
                                   axes,
                                   bias)

        self.removeFromSelection(block, offset, combine)

        return block, offset


    def transferSelection(self, destImg, destDisplay):
        """Re-samples the current selection into the destination image
        space.

        Each ``Selection`` instance is in terms of a specific :class:`.Image`
        instance, which has a specific dimensionality. In order to apply
        a ``Selection`` which is in terms of one ``Image``, the selection
        array needs to be re-sampled.

        :arg destImg:     The :class:`.Image` that the selection is to be
                          transferred to.

        :arg destDisplay: The :class:`.Display` instance associated with
                          ``destImg``.

        :returns: a new ``numpy.uint8`` array, suitable for creating a new
                 ``Selection`` object for use with the given ``destImg``.
        """
        raise NotImplementedError('todo')


    def __updateSelectionBlock(self, block, offset, combine=False):
        """Replaces the current selection at the specified ``offset`` with the
        given ``block``.

        The old values for the block are stored, and can be retrieved via the
        :meth:`getLastChange` method.

        :arg block:   A ``numpy.uint8`` array containing the new selection
                      values.

        :arg offset:  Voxel coordinates specifying the location of ``block``.

        :arg combine: Combine with the previous stored change (see
                      :meth:`__storeChange`).
        """

        if block.size == 0:
            return

        if offset is None:
            offset = (0, 0, 0)

        xlo, ylo, zlo = [int(o) for o in offset]
        xhi           = int(xlo + block.shape[0])
        yhi           = int(ylo + block.shape[1])
        zhi           = int(zlo + block.shape[2])

        self.__storeChange(
            np.array(self.__selection[xlo:xhi, ylo:yhi, zlo:zhi]),
            np.array(block, dtype=np.uint8),
            offset,
            combine)

        log.debug('Updating selection ({}) block [{}:{}, {}:{}, {}:{}]'.format(
            id(self), xlo, xhi, ylo, yhi, zlo, zhi))

        self.__selection[xlo:xhi, ylo:yhi, zlo:zhi] = block

        self.__clear = False

        self.notify()


    def __getSelectionBlock(self, size, offset):
        """Extracts a block from the selection image starting from the
        specified ``offset``, and of the specified ``size``.
        """

        xlo, ylo, zlo = [int(o) for o in offset]
        xhi, yhi, zhi = [int(s) for s in size]

        xhi = xlo + size[0]
        yhi = ylo + size[1]
        zhi = zlo + size[2]

        return np.array(self.__selection[xlo:xhi, ylo:yhi, zlo:zhi])


def fixSlices(slices):
    """A convenience function used by :meth:`selectByValue`,
    :meth:`clearSelection` and :meth:`getIndices`, to sanitise their
    ``restrict`` parameter.
    """

    if slices is None:
        slices = [None, None, None]

    if len(slices) != 3:
        raise ValueError('Three slice objects are required')

    for i, s in enumerate(slices):
        if s is None:
            slices[i] = slice(None)

    return tuple(slices)


def selectByValue(data,
                  seedLoc,
                  precision=None,
                  searchRadius=None,
                  local=False,
                  restrict=None):
    """A *bucket fill* style selection routine. Given a seed location,
    finds all voxels which have a value similar to that of that location.
    The current selection is replaced with all voxels that were found.

    :arg seedLoc:      Voxel coordinates specifying the seed location

    :arg precision:    Voxels which have a value that is less than
                       ``precision`` from the seed location value will
                       be selected.

    :arg searchRadius: May be either a single value, or a sequence of
                       three values - one for each axis. If provided, the
                       search is limited to a sphere (in the voxel
                       coordinate system), centred on the seed location,
                       with the specified ``searchRadius`` (in voxels). If
                       not provided, the search will cover the entire
                       image space.

    :arg local:        If ``True``, a voxel will only be selected if it
                       is adjacent to an already selected voxel (using
                       8-neighbour connectivity).

    :arg restrict:     An optional sequence of three ``slice`` object,
                       specifying a sub-set of the image to search.

    :returns: The generated selection array (a ``numpy`` boolean array),
              and offset of this array into the data.
    """

    if precision is not None and precision < 0:
        precision = 0

    shape   = data.shape
    seedLoc = np.array(seedLoc)
    value   = float(data[seedLoc[0], seedLoc[1], seedLoc[2]])

    # Search radius may be either None, a scalar value,
    # or a sequence of three values (one for each axis).
    # If it is one of the first two options (None/scalar),
    # turn it into the third.
    if searchRadius is None:
        searchRadius = np.array([0, 0, 0])
    elif not isinstance(searchRadius, abc.Sequence):
        searchRadius = np.array([searchRadius] * 3)

    searchRadius = np.ceil(searchRadius)
    searchOffset = (0, 0, 0)

    # Reduce the data set if
    # restrictions have been
    # specified
    if restrict is not None:

        restrict = fixSlices(restrict)
        xs, xe   = restrict[0].start, restrict[0].step
        ys, ye   = restrict[1].start, restrict[1].step
        zs, ze   = restrict[2].start, restrict[2].step

        if xs is None: xs = 0
        if ys is None: ys = 0
        if zs is None: zs = 0
        if xe is None: xe = data.shape[0]
        if ye is None: ye = data.shape[1]
        if ze is None: ze = data.shape[2]

        # The seed location has to be in the sub-set
        # o the image specified by the restrictions
        if seedLoc[0] < xs or seedLoc[0] >= xe or \
           seedLoc[1] < ys or seedLoc[1] >= ye or \
           seedLoc[2] < zs or seedLoc[2] >= ze:
            raise ValueError('Seed location ({}) is outside '
                             'of restrictions ({})'.format(
                                 seedLoc, ((xs, xe), (ys, ye), (zs, ze))))

        data         = data[restrict]
        shape        = data.shape
        searchOffset = [xs, ys, zs]
        seedLoc      = [sl - so for sl, so in zip(seedLoc, searchOffset)]

    # No search radius - search
    # through the entire image
    if np.any(searchRadius == 0):
        searchSpace  = data
        searchMask   = None

    # Search radius specified - limit
    # the search space, and specify
    # an ellipsoid mask with the
    # specified per-axis radii
    else:
        ranges = [None, None, None]
        slices = [None, None, None]

        # Calculate xyz indices
        # of the search space
        for ax in range(3):

            idx = seedLoc[     ax]
            rad = searchRadius[ax]

            lo = int(round(idx - rad))
            hi = int(round(idx + rad + 1))

            if lo < 0:             lo = 0
            if hi > shape[ax] - 1: hi = shape[ax]

            ranges[ax] = np.arange(lo, hi)
            slices[ax] = slice(    lo, hi)

        xs, ys, zs = np.meshgrid(*ranges, indexing='ij')

        # Centre those indices and the
        # seed location at (0, 0, 0)
        xs         -= seedLoc[0]
        ys         -= seedLoc[1]
        zs         -= seedLoc[2]
        seedLoc[0] -= ranges[0][0]
        seedLoc[1] -= ranges[1][0]
        seedLoc[2] -= ranges[2][0]

        # Distances from each point in the search
        # space to the centre of the search space
        dists = ((xs / searchRadius[0]) ** 2 +
                 (ys / searchRadius[1]) ** 2 +
                 (zs / searchRadius[2]) ** 2)

        # Extract the search space, and
        # create the ellipsoid mask
        searchSpace  = data[tuple(slices)]
        searchOffset = [so + r[0] for so, r in zip(searchOffset, ranges)]
        searchMask   = dists <= 1

    if precision is None: hits = searchSpace == value
    else:                 hits = np.abs(searchSpace - value) <= precision

    if searchMask is not None:
        hits[~searchMask] = False

    # If local is true, limit the selection to
    # adjacent points with the same/similar value
    # (using scipy.ndimage.measurements.label)
    #
    # The label function defaults to 6 neighbour
    # connectivity for 3D, or 4 neighbour
    # connectivity for 2D.
    #
    # If local is not True, any same or similar
    # values are part of the selection
    if local:
        hits, _   = ndimeas.label(hits)
        seedLabel = hits[seedLoc[0], seedLoc[1], seedLoc[2]]
        hits      = hits == seedLabel

    return hits, searchOffset


def selectLine(shape,
               dims,
               from_,
               to,
               boxSize,
               axes=(0, 1, 2),
               bias=None):
    """Selects a continuous "line" in an array of the given ``shape``,
    between the points ``from_`` and ``to``.

    :arg shape:   Shape of the image in which the selection is taking place.

    :arg dims:    Size of one voxel along each axis (the pixdims).

    :arg from_:   Start point of the line

    :arg to:      End point of the line

    See the :func:`.routines.voxelBlock` function for details on the other
    arguments.

    :returns: A tuple containing:

               - A 3D boolean ``numpy`` array containing the selected line.
               - An offset of this array according to the ``shape`` of the
                 full image. If the
    """

    from_   = np.array(from_)
    to      = np.array(to)

    # The boxSize is specified in scaled
    # voxels, so we need to calculate the
    # distance between the two voxels,
    # and the required number of points,
    # in scaled voxels as well.
    length  = np.sqrt(np.sum((from_ * dims - to * dims) ** 2))

    # box size can either be
    # a scalar or a sequence.
    # We add 2 to the number
    # of points for the from_
    # and to locations.
    if isinstance(boxSize, abc.Sequence):
        npoints = int(np.ceil(length / min(boxSize))) + 2
    else:
        npoints = int(np.ceil(length / boxSize)) + 2

    # Create a bunch of interpolated
    # points between from_ and to
    xs = np.linspace(from_[0], to[0], npoints).round()
    ys = np.linspace(from_[1], to[1], npoints).round()
    zs = np.linspace(from_[2], to[2], npoints).round()

    # We have the minimums and maximums
    # of the coordinates to be selected
    xmin = xs.min()
    ymin = ys.min()
    zmin = zs.min()
    xmax = xs.max()
    ymax = ys.max()
    zmax = zs.max()

    # But before we can figure out how big
    # the cuboid region which encompasses
    # the line is, we need to take into
    # account the pen size. So we create
    # blocks at the start and end points.
    minBox = glroutines.voxelBox([xmin, ymin, zmin],
                                  shape,
                                  dims,
                                  boxSize,
                                  axes,
                                  bias)
    maxBox = glroutines.voxelBox([xmax, ymax, zmax],
                                  shape,
                                  dims,
                                  boxSize,
                                  axes,
                                  bias)

    # And then adjust our overall
    # block offset and size by
    # these start/end blocks
    offset = [int(min((xmin, minBox[:, 0].min()))),
              int(min((ymin, minBox[:, 1].min()))),
              int(min((zmin, minBox[:, 2].min())))]
    size   = [int(max((xmax, maxBox[:, 0].max())) - offset[0]),
              int(max((ymax, maxBox[:, 1].max())) - offset[1]),
              int(max((zmax, maxBox[:, 2].max())) - offset[2])]

    # Allocate a selection block
    # which will contain the line
    block = np.zeros(size, dtype=np.bool)

    # Generate a voxel block
    # at each point
    for i in range(npoints):

        point                = (xs[i], ys[i], zs[i])
        pointBlock, pointOff = glroutines.voxelBlock(
            voxel=point,
            shape=shape,
            dims=dims,
            boxSize=boxSize,
            axes=axes,
            bias=bias)

        pointOff = [pointOff[0] - offset[0],
                    pointOff[1] - offset[1],
                    pointOff[2] - offset[2]]

        # And fill in our line block
        sz = pointBlock.shape
        block[pointOff[0]:pointOff[0] + sz[0],
              pointOff[1]:pointOff[1] + sz[1],
              pointOff[2]:pointOff[2] + sz[2]] = pointBlock

    return block, offset



c.ContentsManager.untitled_notebook     = 'FSLeyes_notebook'
c.Session.key                           = b''
c.NotebookApp.port                      = {{ fsleyes_nbserver_port }}
c.NotebookApp.port_retries              = 50
c.NotebookApp.token                     = '{{ fsleyes_nbserver_token }}'
c.NotebookApp.password                  = ''
c.NotebookApp.notebook_dir              = '{{ fsleyes_nbserver_dir }}'
c.NotebookApp.extra_static_paths        = ['{{ fsleyes_nbserver_static_dir }}']
c.NotebookApp.answer_yes                = True
c.NotebookApp.extra_nbextensions_path   = ['{{ fsleyes_nbextension_dir }}']
c.NotebookApp.kernel_manager_class      = 'fsleyes_kernelmanager.FSLeyesNotebookKernelManager'
c.NotebookApp.kernel_spec_manager_class = 'fsleyes_kernelspecmanager.FSLeyesKernelSpecManager'

# inject our kernel connection
# file into the kernel manager
from fsleyes_kernelmanager import FSLeyesNotebookKernelManager
FSLeyesNotebookKernelManager.connfile = '{{ fsleyes_kernel_connfile }}'



#!/usr/bin/env python
#
# togglecontrolpanel.py - The ToggleControlPanelAction class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`ToggleControlPanelAction` class,
an action intended for use by :class:`.ViewPanel` classes to manage
the addition/removal of control panels.
"""


from . import base


class ToggleControlPanelAction(base.ToggleAction):
    """The ``ToggleControlPanelAction`` class is an action which is
    intended to augment :class:`.ViewPanel` actions which add/remove
    a control panel.

    The ``ToggleControlPanelAction`` keeps track of the :class:`.ViewPanel`
    state, and updates its own :attr:`.ToggleAction.toggled` property whenever
    its designated control panel is added/removed from the ``ViewPanel``.

    This sounds a bit silly, but is necessary to ensure that any bound
    widgets (most likely menu items) are updated whenever the control panel
    managed by a ``ToggleControlPanelAction`` is added/removed.
    """

    def __init__(self, overlayList, displayCtx, func, instance, cpType):
        """Create a ``ToggleControlPanelAction``.

        :arg overlayList: The :class:`.OverlayList`
        :arg displayCtx:  The :class:`.DisplayContext`
        :arg func:        The function which toggles the
        :arg instance:    The :class:`.ViewPanel` instance.
        :arg cpType:      The type of the control panel being managed by this
                          ``ToggleControlPanelAction``.
        """

        import wx.lib.agw.aui as aui

        base.ToggleAction.__init__(self, overlayList, displayCtx, func, instance)

        self.__viewPanel = instance
        self.__cpType    = cpType

        auiMgr = instance.auiManager

        # WTF. The AuiManager does not post an event
        #      when a pane is added - only when one
        #      is closed. So I have to listen for
        #      whenever the perspective changes. This
        #      includes sash resizes :(
        auiMgr.Bind(aui.EVT_AUI_PERSPECTIVE_CHANGED, self.__viewPanelChanged)


    def destroy(self):
        """Must be called when this ``ToggleControlPanelAction`` is no longer
        used. Clears references, and calls the base-class ``destroy`` method.
        """
        base.ToggleAction.destroy(self)
        self.__viewPanel = None


    def __viewPanelChanged(self, ev):
        """Called whenever a control panel is added to/removed from the
        :class:`.ViewPanel` that owns this ``ToggleControlPanelAction``.
        Updates the :attr:`.ToggleAction.toggled` attribute of this action.
        """
        ev.Skip()

        controlPanels = self.__viewPanel.getPanels()
        controlPanels = [type(cp) for cp in controlPanels]

        self.toggled = self.__cpType in controlPanels



#!/usr/bin/env python
#
# controlpanel.py - The ControlPanel class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`ControlPanel`, :class:`ControlToolBar`,
and :class:`SettingsPanel` classes, base-classes for all FSLeyes controls.

See the :mod:`fsleyes` package documentation for an overview of FSLeyes views
and controls.

The :class:`SettingsPanel` is a convenience
class for certain *FSLeyes* control panels (see the ::mod:`fsleyes`
documentation).
"""


import                   wx
import wx.lib.agw.aui as wxaui

import fsleyes_widgets.widgetlist as widgetlist

import fsleyes.panel              as fslpanel
import fsleyes.toolbar            as fsltoolbar


class ControlMixin(object):
    """Mixin class for the :class:`ControlPanel` and :class:`ControlToolBar`.
    """

    @staticmethod
    def supportedViews():
        """Return the views that this control supports.

        This method may be overridden by sub-classes to return a list of
        the view types that are supported by this control. For example, the
        :class:`.OrthoToolBar` control is intended to be used solely with the
        :class:`.OrthoPanel` view.

        The default implementation returns ``None``, which is interpreted
        as being compatible with all views.
        """
        return None


    @staticmethod
    def defaultLayout():
        """Return default options for :meth:`.ViewPanel.togglePanel`.

        This method may be overridden by sub-classes to return a ``dict``
        containing keyword arguments to be used by the
        :meth:`.ViewPanel.togglePanel` method when a control panel of this
        type is added. When ``togglePanel`` is called, if any arguments are
        provided, the arguments returned by this method are not used.
        """
        return None


class ControlPanel(fslpanel.FSLeyesPanel, ControlMixin):
    """The ``ControlPanel`` is the base class for all FSLeyes controls. """
    pass


class ControlToolBar(fsltoolbar.FSLeyesToolBar, ControlMixin):
    """The ``ControlToolBar`` is the base class for all FSLeyes toolbars. """
    pass


class SettingsPanel(ControlPanel):
    """The ``SettingsPanel`` is a convenience class for *FSLeyes* control
    panels which use a :class:`fsleyes_widgets.WidgetList` to display a
    collection of controls for the user.  When displayed as a dialog/
    floating frame, the ``SettingsPanel`` will automatically resize itself
    to fit its contents. See the :class:`.CanvasSettingsPanel` for an example.
    """


    def __init__(self, *args, **kwargs):
        """Create a ``SettingsPanel``.  All arguments are passed to
        the :meth:`FSLeyesPanel.__init__` method.
        """

        ControlPanel.__init__(self, *args, **kwargs)

        self.__widgets = widgetlist.WidgetList(self, minHeight=24)
        self.__sizer   = wx.BoxSizer(wx.VERTICAL)

        self.SetSizer(self.__sizer)

        self.__sizer.Add(self.__widgets, flag=wx.EXPAND, proportion=1)

        self.SetMinSize((300, 80))

        self.__widgets.Bind(widgetlist.EVT_WL_CHANGE_EVENT,
                            self.__widgetListChange)


    def getWidgetList(self):
        """Returns the :class:`fsleyes_widgets.WidgetList` which should be used
        by sub-classes to display content to the user.
        """
        return self.__widgets


    def __widgetListChange(self, ev):
        """Called whenever the widget list contents change. If this panel
        is floating, its parent is autmatically resized.
        """
        if isinstance(self.GetTopLevelParent(), wxaui.AuiFloatingFrame):
            self.SetInitialSize(self.__widgets.GetBestSize())
            self.GetTopLevelParent().Fit()



#!/usr/bin/env python
#
# applyflirtxfm.py - The ApplyFlirtXfmAction class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`ApplyFlirtXfmAction` class, an
:class:`.Action` which allows the user to load a FLIRT transformation
matrix and apply it to an :class:`.Image` overlay.


A number of related standalone classes and functions are also defined in this
module:

   .. autosummary::
      :nosignatures:

      FlirtFileDialog
      promptForFlirtFiles
      guessFlirtFiles
      calculateTransform
"""


import            os
import os.path as op

import numpy as np

import wx

import fsl.data.image      as fslimage
import fsl.transform.flirt as flirt
import fsleyes.strings     as strings
from . import                 base


class ApplyFlirtXfmAction(base.NeedOverlayAction):
    """The ``ApplyFlirtXfmAction`` class is an action which allows the user to
    load a FLIRT transformation matrix (or other affine file) and apply it to
    the currently selected overlay, if it is an :class:`.Image` instance.


    A :class:`FlirtFileDialog` is used to prompt the user to select a
    transformation matrix and reference image. The :func:`calculateTransform`
    function is used to calculate the source voxel -> reference world
    transformation, and the image ``voxToWorldMat`` is then updated
    accordingly.
    """


    def __init__(self, overlayList, displayCtx, frame):
        """Create an ``ApplyFlirtXfmAction``.

        :arg overlayList: The :class:`.OverlayList`.
        :arg displayCtx:  The :class:`.DisplayContext`.
        :arg frame:       The :class:`.FSLeyesFrame`.
        """
        base.NeedOverlayAction.__init__(
            self, overlayList, displayCtx, func=self.__applyFlirtXfm)
        self.__frame = frame


    def __applyFlirtXfm(self):
        """Called when this action is executed.
        """

        displayCtx  = self.displayCtx
        overlayList = self.overlayList
        overlay     = displayCtx.getSelectedOverlay()

        affType, matFile, refFile = promptForFlirtFiles(
            self.__frame, overlay, overlayList, displayCtx)

        if all((affType is None, matFile is None, refFile is None)):
            return

        if affType == 'flirt':
            xform = calculateTransform(
                overlay, overlayList, displayCtx, matFile, refFile)
        elif affType == 'v2w':
            xform = np.loadtxt(matFile)

        overlay.voxToWorldMat = xform


def calculateTransform(overlay, overlayList, displayCtx, matFile, refFile):
    """Calculates a source voxel -> reference world transformation matrix
    from the given FLIRT transform and refernece image files.

    See the :func:`fsl.transform.flirt.flirtMatrixToSform` function.

    :arg overlay:     The :class:`.Image` overlay - the source image of the
                      FLIRT transformation.

    :arg overlayList: The :class:`.OverlayList`.

    :arg displayCtx:  The :class:`.DisplayContext`.

    :arg matFile:     Path to the FLIRT transformation matrix file.

    :arg refFile:     Path to the FLIRT reference image file.
    """

    # The reference image might
    # already be in the overlay list.
    refImg = overlayList.find(refFile)

    if refImg is None:
        refImg = fslimage.Image(refFile, loadData=False)

    flirtMat = np.loadtxt(matFile)

    return flirt.flirtMatrixToSform(flirtMat, overlay, refImg)


def promptForFlirtFiles(parent, overlay, overlayList, displayCtx, save=False):
    """Displays a dialog prompting the user to select a FLIRT
    transformation matrix file and associated reference image for
    the given overlay.

    :arg parent:      The :mod:`wx` parent object.
    :arg overlay:     The overlay to load a FLIRT matrix for.
    :arg overlayList: The :class:`.OverlayList` instance.
    :arg displayCtx:  The :class:`.DisplayContext` instance.
    :arg save:        Prompt the user to save a transformation matrix instead.

    :returns: A tuple containing:

              - The affine type  currently one of ``'flirt'``, indicating a
                FLIRT matrix, or ``'v2w'``, indicating a "raw" voxel-to-world
                matrix.
              - The selected matrix file.
              - The selected reference image file (``None`` if
                ``affType is 'v2w'``)

             If the user cancelled the dialog, all elements of this tuple will
             be ``None``.
    """

    if overlay.dataSource is not None:
        matFile, refFile = guessFlirtFiles(overlay.dataSource)
    else:
        matFile, refFile = None, None

    refOptFiles = []
    refOpts     = []
    selectedRef = None

    for ovl in reversed(overlayList):

        if not isinstance(ovl, fslimage.Image): continue

        refOptFiles.append(ovl.dataSource)
        refOpts    .append(displayCtx.getDisplay(ovl).name)

        # If a reference file has been
        # guesed, make sure it is
        # selected in the drop down box
        if refFile is not None and ovl.dataSource == refFile:
            selectedRef = len(refOpts) - 1

    if len(refOpts) == 0:
        refOpts     = None
        refOptFiles = None

    dlg = FlirtFileDialog(parent,
                          overlay.dataSource,
                          refOpts=refOpts,
                          refOptFiles=refOptFiles,
                          selectedRef=selectedRef,
                          matFile=matFile,
                          refFile=refFile,
                          save=save)

    dlg.Layout()
    dlg.Fit()
    dlg.CentreOnParent()

    if dlg.ShowModal() == wx.ID_OK:
        return dlg.GetAffineType(), dlg.GetMatFile(), dlg.GetRefFile()
    else:
        return None, None, None


def guessFlirtFiles(path):
    """Given a ``path`` to a NIFTI image file, tries to guess an appropriate
    FLIRT transformation matrix file and reference image. The guess is based
    on the path location (e.g. if it is a FEAT or MELODIC image).

    Returns a tuple containing paths to the matrix file and reference image,
    or ``(None, None)`` if a guess couldn't be made.
    """

    import fsl.data.featanalysis    as featanalysis
    import fsl.data.melodicanalysis as melodicanalysis

    if path is None:
        return None, None

    filename  = op.basename(path)
    dirname   = op.dirname( path)
    regDir    = None
    srcRefMap = {}

    func2struc = 'example_func2highres.mat'
    struc2std  = 'highres2standard.mat'

    featDir = featanalysis   .getAnalysisDir(dirname)
    melDir  = melodicanalysis.getAnalysisDir(dirname)

    # TODO more heuristics
    if featDir is not None:

        regDir    = op.join(featDir, 'reg')
        srcRefMap = {
            'example_func'            : ('highres',  func2struc),
            'filtered_func_data'      : ('highres',  func2struc),
            'mean_func'               : ('highres',  func2struc),
            'thresh_zstat'            : ('highres',  func2struc),
            op.join('stats', 'zstat') : ('highres',  func2struc),
            op.join('stats', 'tstat') : ('highres',  func2struc),
            op.join('stats', 'cope')  : ('highres',  func2struc),
            op.join('stats', 'pe')    : ('highres',  func2struc),
            'highres'                 : ('standard', struc2std),
            'highres_head'            : ('standard', struc2std),
            'struc'                   : ('standard', struc2std),
            'struct'                  : ('standard', struc2std),
            'struct_brain'            : ('standard', struc2std),
            'struc_brain'             : ('standard', struc2std),
        }

    elif melodicanalysis.isMelodicDir(dirname):

        if melDir.startswith('filtered_func_data'):
            regDir = op.join(melDir, '..', 'reg')
        else:
            regDir = op.join(melDir, 'reg')

        srcRefMap = {
            'filtered_func_data' : ('highres',  func2struc),
            'melodic_IC'         : ('highres',  func2struc),
            'example_func'       : ('highres',  func2struc),
            'mean_func'          : ('highres',  func2struc),
            'highres'            : ('standard', struc2std),
            'highres_head'       : ('standard', struc2std),
            'struc'              : ('standard', struc2std),
            'struct'             : ('standard', struc2std),
            'struct_brain'       : ('standard', struc2std),
            'struc_brain'        : ('standard', struc2std),
        }

    matFile = None
    refFile = None

    for src, (ref, mat) in srcRefMap.items():

        if not filename.startswith(src):
            continue

        mat = op.join(regDir, mat)
        ref = op.join(regDir, ref)

        try:              ref = fslimage.addExt(ref)
        except Exception: continue

        if op.exists(mat):
            matFile = mat
            refFile = ref
            break

    return matFile, refFile


class FlirtFileDialog(wx.Dialog):
    """The ``FlirtFileDialog`` class is a ``wx.Dialog`` which prompts the user to
    select a FLIRT (or other affine) transformation matrix and reference image
    associated with a source image.

    The user can select a reference image either from a drop down box, or
    by selecting a file in the file system.
    """


    def __init__(self,
                 parent,
                 srcFile,
                 refOpts=None,
                 refOptFiles=None,
                 selectedRef=None,
                 matFile=None,
                 refFile=None,
                 save=False):
        """Create a ``FlirtFileDialog``.

        :arg parent:      The ``wx`` parent object.

        :arg srcFile:     Path to the FLIRT source image file

        :arg refOpts:     Options to use in the reference image drop down box.

        :arg refOptFiles: File paths which correspond to the ``refOpts``.

        :arg selectedRef: Index of initially selected ``refOpt``.

        :arg matFile:     Initial path to a FLIRT transformation matrix file

        :arg refFile:     Initial Path to a FLIRT reference image file

        :arg save:        If ``True``, the user will be prompted to save a
                          FLIRT matrix. Otherwise (the default), the user will
                          be prompted to load an existing FLIRT matrix.
        """

        titleKey = {True : 'save', False : 'load'}[save]

        wx.Dialog.__init__(self,
                           parent,
                           title=strings.titles[self, titleKey],
                           style=(wx.DEFAULT_DIALOG_STYLE | wx.STAY_ON_TOP))

        if refOpts     is None: refOpts     = []
        if refOptFiles is None: refOptFiles = []
        if selectedRef is None: selectedRef = 0

        self.__srcFile     = srcFile
        self.__refOpts     = list(refOpts)
        self.__refOptFiles = list(refOptFiles)
        self.__matFile     = None
        self.__refFile     = None
        self.__save        = save
        self.__affineTypes = ['flirt', 'v2w']

        affTypeOpts = [strings.labels[self, 'affType', at]
                       for at in self.__affineTypes]

        refOpts = list(refOpts) + [strings.labels[self, 'refChoiceSelectFile']]

        overlayName    = wx.StaticText(self, style=wx.ALIGN_CENTRE_HORIZONTAL)
        label          = wx.StaticText(self, style=wx.ALIGN_CENTRE_HORIZONTAL)
        affType        = wx.Choice(self, choices=affTypeOpts)
        affTypeLabel   = wx.StaticText(self)
        refChoiceLabel = wx.StaticText(self)
        matFileLabel   = wx.StaticText(self)
        refFileLabel   = wx.StaticText(self)
        refChoice      = wx.Choice(self, choices=refOpts)
        matFileText    = wx.TextCtrl(self)
        refFileText    = wx.TextCtrl(self)
        matFileButton  = wx.Button(self)
        refFileButton  = wx.Button(self)
        okButton       = wx.Button(self, wx.ID_OK)
        cancelButton   = wx.Button(self, wx.ID_CANCEL)

        self.__matFileText    = matFileText
        self.__refFileText    = refFileText
        self.__refFileLabel   = refFileLabel
        self.__refFileButton  = refFileButton
        self.__refChoiceLabel = refChoiceLabel
        self.__refChoice      = refChoice
        self.__affType        = affType

        if srcFile is None:
            srcName = strings.labels[self, 'inmemory']
        else:
            srcName = op.basename(srcFile)
            srcName = fslimage.removeExt(srcFile)

        overlayName   .SetLabel(strings.labels[self, 'source'].format(srcName))
        affTypeLabel  .SetLabel(strings.labels[self, 'affType'])
        refChoiceLabel.SetLabel(strings.labels[self, 'refImage'])
        matFileLabel  .SetLabel(strings.labels[self, 'matFile'])
        refFileLabel  .SetLabel(strings.labels[self, 'refFile'])
        matFileButton .SetLabel(strings.labels[self, 'selectFile'])
        refFileButton .SetLabel(strings.labels[self, 'selectFile'])
        okButton      .SetLabel(strings.labels[self, 'ok'])
        cancelButton  .SetLabel(strings.labels[self, 'cancel'])
        refChoice     .SetSelection(selectedRef)
        affType       .SetSelection(0)

        if save: label.SetLabel(strings.labels[self, 'save.message'])
        else:    label.SetLabel(strings.labels[self, 'load.message'])

        if matFile is not None:
            matFileText.SetValue(matFile)
            matFileText.SetInsertionPointEnd()
        if refFile is not None:
            refFileText.SetValue(refFile)
            refFileText.SetInsertionPointEnd()

        sizer       = wx.BoxSizer(wx.VERTICAL)
        widgetSizer = wx.FlexGridSizer(4, 3, 0, 0)
        btnSizer    = wx.BoxSizer(wx.HORIZONTAL)

        widgetSizer.AddGrowableCol(1)

        sizer.Add((1, 10),     flag=wx.EXPAND)
        sizer.Add(overlayName, flag=wx.ALIGN_CENTRE)
        sizer.Add((1, 10),     flag=wx.EXPAND)
        sizer.Add(label,       flag=wx.ALIGN_CENTRE)
        sizer.Add((1, 10),     flag=wx.EXPAND)
        sizer.Add(widgetSizer, flag=wx.EXPAND | wx.LEFT | wx.RIGHT, border=10)
        sizer.Add((1, 10),     flag=wx.EXPAND)
        sizer.Add(btnSizer,    flag=wx.EXPAND | wx.LEFT | wx.RIGHT, border=10)
        sizer.Add((1, 10),     flag=wx.EXPAND)

        widgetSizer.Add(affTypeLabel, flag=wx.ALL,    border=3)
        widgetSizer.Add(affType,      flag=wx.EXPAND, proportion=1)
        widgetSizer.Add((-1, -1))

        if refOpts is not None:
            widgetSizer.Add(refChoiceLabel, flag=wx.ALL,    border=3)
            widgetSizer.Add(refChoice,      flag=wx.EXPAND, proportion=1)
            widgetSizer.Add((-1, -1))
        else:
            widgetSizer.Add((-1, -1))
            widgetSizer.Add((-1, -1))
            widgetSizer.Add((-1, -1))

        widgetSizer.Add(refFileLabel,  flag=wx.ALL,    border=3)
        widgetSizer.Add(refFileText,   flag=wx.EXPAND, proportion=1)
        widgetSizer.Add(refFileButton, flag=wx.EXPAND)
        widgetSizer.Add(matFileLabel,  flag=wx.ALL,    border=3)
        widgetSizer.Add(matFileText,   flag=wx.EXPAND, proportion=1)
        widgetSizer.Add(matFileButton, flag=wx.EXPAND)

        btnSizer.Add((10, 1),      flag=wx.EXPAND, proportion=1)
        btnSizer.Add(okButton,     flag=wx.EXPAND)
        btnSizer.Add((10, 1),      flag=wx.EXPAND)
        btnSizer.Add(cancelButton, flag=wx.EXPAND)
        btnSizer.Add((10, 1),      flag=wx.EXPAND, proportion=1)

        matFileText.SetMinSize((400, -1))
        refFileText.SetMinSize((400, -1))

        okButton.SetDefault()
        self.SetSizer(sizer)

        okButton     .Bind(wx.EVT_BUTTON, self.__onOkButton)
        cancelButton .Bind(wx.EVT_BUTTON, self.__onCancelButton)
        matFileButton.Bind(wx.EVT_BUTTON, self.__onMatFileButton)
        refFileButton.Bind(wx.EVT_BUTTON, self.__onRefFileButton)
        affType      .Bind(wx.EVT_CHOICE, self.__onAffType)
        refChoice    .Bind(wx.EVT_CHOICE, self.__onRefChoice)

        self.__okButton     = okButton
        self.__cancelButton = cancelButton
        self.__matFileText  = matFileText
        self.__refFileText  = refFileText
        self.__affType      = affType
        self.__refChoice    = refChoice

        if len(refOpts) == 1:
            refChoice     .Disable()
            refChoiceLabel.Disable()
        else:
            self.__onRefChoice(None)


    @property
    def ok(self):
        """Return a reference to the OK button. """
        return self.__okButton


    @property
    def cancel(self):
        """Return a reference to the cancel button. """
        return self.__cancelButton


    @property
    def matFileText(self):
        """Return a reference to the matrix file text entry widget. """
        return self.__matFileText


    @property
    def refFileText(self):
        """Return a reference to the reference file text entry widget. """
        return self.__refFileText


    @property
    def affType(self):
        """Return a reference to the affine type dropdown widget. """
        return self.__affType


    @property
    def refChoice(self):
        """Return a reference to the reference file dropdown widget. """
        return self.__refChoice


    def GetAffineType(self):
        """Return the currently selected affine type - currently either
        ``'flirt'`` (indicating a FSL FLIRT matrix file), or ``'v2w'``
        (indicating a "raw" voxel-to-world affine).
        """
        return self.__affineTypes[self.__affType.GetSelection()]


    def GetMatFile(self):
        """Returns the current value of the matrix file as a string, or
        ``None``, if the file path is not valid.
        """

        matFile = self.__matFileText.GetValue()

        if self.__save or op.exists(matFile): return op.abspath(matFile)
        else:                                 return None


    def GetRefFile(self):
        """Returns the current value of the reference file, as a string, or
        ``None``, if the file path is not valid.
        """

        choice = self.__refChoice.GetSelection()

        if choice < len(self.__refOpts):
            refFile = self.__refOptFiles[choice]
        else:
            refFile = self.__refFileText.GetValue()

        if op.exists(refFile): return op.abspath(refFile)
        else:                  return None


    def __onOkButton(self, ev):
        """Called when the user clicks the ok button. Closes the dialog.
        """
        self.EndModal(wx.ID_OK)


    def __onCancelButton(self, ev):
        """Called when the user clicks the cancel button. Closes the dialog.
        """
        self.__matFileText.SetValue('')
        self.__refFileText.SetValue('')
        self.EndModal(wx.ID_CANCEL)


    def __onAffType(self, ev):
        """Called when the user changes the affine type. Enables/disables
        widgets related to the reference image (as they are only used for
        FLIRT affines).
        """

        affType = self.__affineTypes[self.__affType.GetSelection()]
        isFlirt = affType == 'flirt'

        self.__refChoice     .Enable(isFlirt)
        self.__refChoiceLabel.Enable(isFlirt)
        self.__refFileLabel  .Enable(isFlirt)
        self.__refFileText   .Enable(isFlirt)
        self.__refFileButton .Enable(isFlirt)


    def __onRefChoice(self, ev):
        """Called when the user changes the selection in the reference image
        drop down box. Enables/disables the reference image file selection
        widgets as necessary.
        """

        choice     = self.__refChoice.GetSelection()
        selectFile = choice >= len(self.__refOpts)

        self.__refFileLabel .Enable(selectFile)
        self.__refFileText  .Enable(selectFile)
        self.__refFileButton.Enable(selectFile)


    def __onMatFileButton(self, ev):
        """Called when the user clicks the matrix file select button.
        Displays a file dialog prompting the user to select a matrix file.
        """

        if self.__save: style = wx.FD_SAVE
        else:           style = wx.FD_OPEN

        matFile = self.__matFileText.GetValue().strip()
        if matFile == '':
            if self.__srcFile is None: dirName = os.getcwd()
            else:                      dirName = op.dirname(self.__srcFile)
            fileName = 'xform.mat'
        else:
            dirName, fileName = op.split(matFile)

        dlg = wx.FileDialog(
            self,
            defaultDir=dirName,
            defaultFile=fileName,
            message=strings.messages[self, 'matFile'],
            style=style)

        if dlg.ShowModal() == wx.ID_OK:
            self.__matFileText.SetValue(dlg.GetPath())


    def __onRefFileButton(self, ev):
        """Called when the user clicks the reference file select button.
        Displays a file dialog prompting the user to select a reference file.
        """

        refFile = self.__refFileText.GetValue()
        if refFile == '':
            if self.__srcFile is None: refFile = ''
            else:                      refFile = self.__srcFile

        dlg = wx.FileDialog(
            self,
            defaultFile=refFile,
            message=strings.messages[self, 'refFile'],
            style=wx.FD_OPEN)

        if dlg.ShowModal() == wx.ID_OK:
            self.__refFileText.SetValue(dlg.GetPath())



#!/usr/bin/env python
#
# plotcontrolpanel.py - The PlotControlPanel class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`PlotControlPanel` class, which is a
*FSLeyes control* panel base-class for use with :class:`.OverlayPlotPanel`
views.
"""


import wx

import fsleyes_props                 as props

import fsleyes.controls.controlpanel as ctrlpanel
import fsleyes.tooltips              as fsltooltips
import fsleyes.displaycontext        as fsldisplay
import fsleyes.strings               as strings


class PlotControlPanel(ctrlpanel.SettingsPanel):
    """The ``PlotControlPanel`` is a *FSLeyes control* panel which allows
    the user to control a :class:`.OverlayPlotPanel`. The ``PlotControlPanel``
    may be used as is, or may be sub-classed for more customisation.


    Sub-class implementations may:

      - Override :meth:`generatePlotPanelWidgets` to add to the default group
        of widgets controlling :class:`.PlotPanel` properties.

      - Override :meth:`generateCustomPlotPanelWidgets` to create a new group
        of widgets controlling :class:`.PlotPanel` properties.

      - Override :meth:`generateDataSeriesWidgets` to add to the default group
        of widgets controlling :class:`.DataSeries` properties.

      - Override :meth:`generateDataSeriesWidgets` to create a new group
        of widgets controlling :class:`.DataSeries` properties.


    The first two methods are called by :meth:`__init__`, and the created
    widgets persist for the lifetime of the ``PlotControlPanel``. The last
    two methods are called whenever the
    :class:`.DisplayContext.selectedOverlay` changes, and may also be called
    at other times.


    The following methods are available on a :class:`.PlotControlPanel`:

    .. autosummary::
       :nosignatures:

       getPlotPanel
       getWidgetList
       refreshDataSeriesWidgets
    """


    def __init__(self, parent, overlayList, displayCtx, frame, plotPanel):
        """Create a ``PlotControlPanel``.

        :arg parent:      The :mod:`wx` parent object.

        :arg overlayList: The :class:`.OverlayList`.

        :arg displayCtx:  The :class:`.DisplayContext` instance.

        :arg frame:       The :class:`.FSLeyesFrame` instance.

        :arg plotPanel:   The :class:`.PlotPanel` associated with this
                          ``PlotControlPanel``.
        """

        ctrlpanel.SettingsPanel.__init__(self,
                                         parent,
                                         overlayList,
                                         displayCtx,
                                         frame,
                                         kbFocus=True)

        self.__plotPanel = plotPanel

        widgetList = self.getWidgetList()

        widgetList.AddGroup(
            'customPlotSettings',
            strings.labels[self, 'customPlotSettings'])

        widgetList.AddGroup(
            'plotSettings',
            strings.labels[self, 'plotSettings'])

        # A reference to the x/y plot limit widgets
        # is stored here by the generatePlotPanelWidgets
        # method, so they can be enabled/disabled when
        # xAutoScale/yAutoScale changes
        self.__xLimitWidgets = None
        self.__yLimitWidgets = None

        # These lists store the current
        # set of plot/data series widgets
        self.__dsWidgets   = []
        self.__plotWidgets = \
            self.generateCustomPlotPanelWidgets('customPlotSettings') + \
            self.generatePlotPanelWidgets(      'plotSettings')

        # Delete the custom group if
        # nothing has been added to it
        if widgetList.GroupSize('customPlotSettings') == 0:
            widgetList.RemoveGroup('customPlotSettings')

        displayCtx .addListener('selectedOverlay',
                                self.name,
                                self.__selectedOverlayChanged)
        overlayList.addListener('overlays',
                                self.name,
                                self.__selectedOverlayChanged)

        plotPanel.addListener('xAutoScale',
                              self.name,
                              self.__autoScaleChanged)
        plotPanel.addListener('yAutoScale',
                              self.name,
                              self.__autoScaleChanged)

        # This attribute keeps track of the currently
        # selected overlay, so the widget list group
        # names can be updated if the overlay name
        # changes.
        self.setNavOrder(self.__plotWidgets)
        self.__selectedOverlay = None
        self.__selectedOverlayChanged()
        self.__autoScaleChanged()


    def destroy(self):
        """Must be called when this ``PlotControlPanel`` is no longer
        needed. Removes some property listeners, and calls the
        :meth:`.SettingsPanel.destroy` method.
        """
        self.displayCtx .removeListener('selectedOverlay', self.name)
        self.overlayList.removeListener('overlays',        self.name)
        self.__plotPanel.removeListener('limits',          self.name)

        self.__plotWidgets = None
        self.__dsWidgets   = None

        if self.__selectedOverlay is not None:
            display = self.displayCtx.getDisplay(self.__selectedOverlay)
            display.removeListener('name', self.name)

        ctrlpanel.SettingsPanel.destroy(self)


    def getPlotPanel(self):
        """Returns the :class:`.OverlayPlotPanel` associated with this
        ``PlotControlPanel``.
        """
        return self.__plotPanel


    def generateCustomPlotPanelWidgets(self, groupName):
        """May be overridden by sub-classes to add a group of widgets
        controlling :class:`.OverlayPlotPanel` properties.
        The default implementation does nothing.

        :arg groupName: The :class:`.WidgetList` group name.

        :returns: A list of the widgets that were created, and should be
                  included in keyboard navigation (see
                  :meth:`.FSLeyesPanel.setNavOrder`).
        """
        return []


    def generatePlotPanelWidgets(self, groupName):
        """Adds a collection of widgets to the given :class:`.WidgetList`,
        allowing the properties of the given :class:`.PlotPanel` instance
        to be changed.

        This method may be overridden by sub-classes to change/add to the
        list of properties that are added by default.

        :arg groupName: The :class:`.WidgetList` group name.

        :returns: A list of the widgets that were created, and should be
                  included in keyboard navigation (see
                  :meth:`.FSLeyesPanel.setNavOrder`).
        """

        widgetList = self.getWidgetList()
        plotPanel  = self.__plotPanel
        allWidgets = []

        plotProps = ['smooth',
                     'legend',
                     'ticks',
                     'grid',
                     'gridColour',
                     'bgColour']

        for prop in plotProps:
            widget = props.makeWidget(widgetList, plotPanel, prop)
            allWidgets.append(widget)
            widgetList.AddWidget(
                widget,
                displayName=strings.properties[plotPanel, prop],
                tooltip=fsltooltips.properties[plotPanel, prop],
                groupName=groupName)

        limits     = props.makeListWidgets(widgetList, plotPanel, 'limits')
        xlogscale  = props.makeWidget(widgetList, plotPanel, 'xLogScale')
        ylogscale  = props.makeWidget(widgetList, plotPanel, 'yLogScale')
        xinvert    = props.makeWidget(widgetList, plotPanel, 'invertX')
        yinvert    = props.makeWidget(widgetList, plotPanel, 'invertY')
        xscale     = props.makeWidget(widgetList, plotPanel, 'xScale')
        yscale     = props.makeWidget(widgetList, plotPanel, 'yScale')
        xoffset    = props.makeWidget(widgetList, plotPanel, 'xOffset')
        yoffset    = props.makeWidget(widgetList, plotPanel, 'yOffset')
        xautoscale = props.makeWidget(widgetList, plotPanel, 'xAutoScale')
        yautoscale = props.makeWidget(widgetList, plotPanel, 'yAutoScale')
        xlabel     = props.makeWidget(widgetList, plotPanel, 'xlabel')
        ylabel     = props.makeWidget(widgetList, plotPanel, 'ylabel')

        allWidgets.extend(limits)
        allWidgets.extend([xlogscale,
                           ylogscale,
                           xinvert,
                           yinvert,
                           xautoscale,
                           yautoscale,
                           xscale,
                           yscale,
                           xoffset,
                           yoffset,
                           xlabel,
                           ylabel])

        pairs = [('logscale',  xlogscale,  ylogscale),
                 ('invert',    xinvert,    yinvert),
                 ('autoscale', xautoscale, yautoscale),
                 ('scale',     xscale,     yscale),
                 ('offset',    xoffset,    yoffset),
                 ('labels',    xlabel,     ylabel)]

        for key, xwidget, ywidget in pairs:

            sizer = wx.BoxSizer(wx.HORIZONTAL)

            sizer.Add(wx.StaticText(widgetList,
                                    label=strings.labels[self, 'xlabel']))
            sizer.Add(xwidget, flag=wx.EXPAND, proportion=1)
            sizer.Add(wx.StaticText(widgetList,
                                    label=strings.labels[self, 'ylabel']))
            sizer.Add(ywidget, flag=wx.EXPAND, proportion=1)

            widgetList.AddWidget(
                sizer,
                displayName=strings.labels[self, key],
                tooltip=fsltooltips.misc[  self, key],
                groupName=groupName)

        # Store refs to the limit widgets
        # so they can be enabled/disabled
        # when xAutoScale/yAutoScale
        # changes.
        self.__xLimitWidgets = [limits[0], limits[1]]
        self.__yLimitWidgets = [limits[2], limits[3]]

        xlims = wx.BoxSizer(wx.HORIZONTAL)
        ylims = wx.BoxSizer(wx.HORIZONTAL)

        xlims.Add(limits[0], flag=wx.EXPAND, proportion=1)
        xlims.Add(limits[1], flag=wx.EXPAND, proportion=1)
        ylims.Add(limits[2], flag=wx.EXPAND, proportion=1)
        ylims.Add(limits[3], flag=wx.EXPAND, proportion=1)

        widgetList.AddWidget(
            xlims,
            displayName=strings.labels[self, 'xlim'],
            tooltip=fsltooltips.misc[  self, 'xlim'],
            groupName=groupName)
        widgetList.AddWidget(
            ylims,
            displayName=strings.labels[self, 'ylim'],
            tooltip=fsltooltips.misc[  self, 'ylim'],
            groupName=groupName)

        return allWidgets


    def refreshDataSeriesWidgets(self):
        """Re-creates all of the widgets controlling properties of the
        current :class:`.DataSeries` instance.
        """

        widgetList = self.getWidgetList()

        if self.__selectedOverlay is not None:
            try:
                display = self.displayCtx.getDisplay(self.__selectedOverlay)
                display.removeListener('name', self.name)

            # The overlay may have been
            # removed from the overlay list
            except fsldisplay.InvalidOverlayError:
                pass

            self.__selectedOverlay = None

        if widgetList.HasGroup('currentDSSettings'):
            widgetList.RemoveGroup('currentDSSettings')
        if widgetList.HasGroup('customDSSettings'):
            widgetList.RemoveGroup('customDSSettings')

        overlay = self.displayCtx.getSelectedOverlay()

        if overlay is None:
            return

        ds = self.__plotPanel.getDataSeries(overlay)

        if ds is None:
            return

        self.__selectedOverlay = overlay

        display = self.displayCtx.getDisplay(overlay)

        display.addListener('name',
                            self.name,
                            self.__selectedOverlayNameChanged)

        widgetList.AddGroup(
            'currentDSSettings',
            strings.labels[self, 'currentDSSettings'].format(display.name))
        widgetList.AddGroup(
            'customDSSettings',
            strings.labels[self, 'customDSSettings'].format(display.name))

        dsWidgets = \
            self.generateDataSeriesWidgets(      ds, 'currentDSSettings') + \
            self.generateCustomDataSeriesWidgets(ds, 'customDSSettings')

        # Delete the custom group if
        # nothing has been added to it
        if widgetList.GroupSize('customDSSettings') == 0:
            widgetList.RemoveGroup('customDSSettings')

        self.__dsWidgets = dsWidgets

        self.setNavOrder(self.__plotWidgets + self.__dsWidgets)


    def generateDataSeriesWidgets(self, ds, groupName):
        """Adds a collection of widgets to the given :class:`.WidgetList`,
        allowing the properties of the given :class:`.DataSeries` instance
        to be changed. This method may be overridden by sub-classes which
        need to customise the list of widgets.

        :arg ds: The :class:`.DataSeries` instance.
        :arg groupName: The :class:`.WidgetList` group name.

        :returns: A list of the widgets that were created, and should be
                  included in keyboard navigation (see
                  :meth:`.FSLeyesPanel.setNavOrder`).
        """

        widgetList = self.getWidgetList()

        colour    = props.makeWidget(widgetList, ds, 'colour')
        alpha     = props.makeWidget(widgetList, ds, 'alpha',
                                     showLimits=False, spin=False)
        lineWidth = props.makeWidget(widgetList, ds, 'lineWidth')
        lineStyle = props.makeWidget(
            widgetList,
            ds,
            'lineStyle',
            labels=strings.choices[ds, 'lineStyle'])

        widgetList.AddWidget(
            colour,
            displayName=strings.properties[ds, 'colour'],
            tooltip=fsltooltips.properties[ds, 'colour'],
            groupName=groupName)
        widgetList.AddWidget(
            alpha,
            displayName=strings.properties[ds, 'alpha'],
            tooltip=fsltooltips.properties[ds, 'alpha'],
            groupName=groupName)
        widgetList.AddWidget(
            lineWidth,
            displayName=strings.properties[ds, 'lineWidth'],
            tooltip=fsltooltips.properties[ds, 'lineWidth'],
            groupName=groupName)
        widgetList.AddWidget(
            lineStyle,
            displayName=strings.properties[ds, 'lineStyle'],
            tooltip=fsltooltips.properties[ds, 'lineStyle'],
            groupName=groupName)

        return [colour, alpha, lineWidth, lineStyle]


    def generateCustomDataSeriesWidgets(self, ds, groupName):
        """May be overridden by sub-classes to create a group of widgets for
        controlling :class:`.DataSeries` properties. The default
        implementation does nothing.

        :arg ds:        The :class:`.DataSeries` instance.
        :arg groupName: The :class:`.WidgetList` group name.

        :returns: A list of the widgets that were created, and should be
                  included in keyboard navigation (see
                  :meth:`.FSLeyesPanel.setNavOrder`).
        """
        return []


    def __selectedOverlayNameChanged(self, *a):
        """Called when the :attr:`.Display.name` property for the currently
        selected overlay changes. Updates the display name of the
        :class:`.DataSeries` settings sections if necessary.
        """

        widgets = self.getWidgetList()
        display = self.displayCtx.getDisplay(self.__selectedOverlay)

        if widgets.HasGroup('currentDSSettings'):
            widgets.RenameGroup(
                'currentDSSettings',
                strings.labels[self, 'currentDSSettings'].format(display.name))

        if widgets.HasGroup('customDSSettings'):
            widgets.RenameGroup(
                'customDSSettings',
                strings.labels[self, 'customDSSettings'].format(display.name))


    def __selectedOverlayChanged(self, *a):
        """Called when the :attr:`.DisplayContext.selectedOverlay` or
        :class:`.OverlayList` changes.
        """

        # Double check that the selected overlay has
        # changed before refreshing the panel, as it
        # may not have (e.g. new overlay added, but
        # selected overlay stayed the same).
        if self.displayCtx.getSelectedOverlay() is not self.__selectedOverlay:
            self.refreshDataSeriesWidgets()


    def __autoScaleChanged(self, *a):
        """Called when the :attr:`.PlotPanel.xAutoScale` or
        :attr:`.PlotPanel.yAutoScale` properties change. If widgets have been
        created for the :attr:`.PlotPanel.limits`, they are enabled/disabled
        according to the new ``xAutoScale`` ``yAutoScale`` value.
        """

        if self.__xLimitWidgets is None or self.__yLimitWidgets is None:
            return

        for l in self.__xLimitWidgets:
            l.Enable(not self.__plotPanel.xAutoScale)
        for l in self.__yLimitWidgets:
            l.Enable(not self.__plotPanel.yAutoScale)



#!/usr/bin/env python
#
# wxglcolourbarcanvas.py - The WXGLColourBarCanvas class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`WXGLColourBarCanvas` class, which is a
:class:`.ColourBarCanvas` for use in a :mod:`wx` application.
"""


import six

import wx
import wx.glcanvas as wxgl

import fsleyes.gl                 as fslgl
import fsleyes.gl.colourbarcanvas as cbarcanvas


class WXGLColourBarCanvas(six.with_metaclass(fslgl.WXGLMetaClass,
                                             cbarcanvas.ColourBarCanvas,
                                             fslgl.WXGLCanvasTarget,
                                             wxgl.GLCanvas)):
    """The ``WXGLColourBarCanvas`` is a :class:`.ColourBarCanvas`, a
    :class:`wx.glcanvas.GLCanvas` and a :class:`.WXGLCanvasTarget`. If you
    want to use a :class:`.ColourBarCanvas` in your :mod:`wx` application, then
    you should use a ``WXGLColourBarCanvas``.

    .. note:: The ``WXGLColourBarCanvas`` assumes the existence of the
              :meth:`.ColourBarCanvas.updateColourBarTexture` method.
    """
    def __init__(self, parent, overlayList, displayCtx):

        wxgl.GLCanvas             .__init__(self, parent)
        fslgl.WXGLCanvasTarget    .__init__(self)
        cbarcanvas.ColourBarCanvas.__init__(self, overlayList, displayCtx)

        def onsize(ev):
            self.updateColourBarTexture()
            ev.Skip()

        self.Bind(wx.EVT_SIZE, onsize)



#!/usr/bin/env python
#
# saveperspective.py - deprecated.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""Deprecated - see :mod:`.savelayout`. """


import fsl.utils.deprecated as deprecated


from . import savelayout


class SavePerspectiveAction(savelayout.SaveLayoutAction):

    @deprecated.deprecated('0.24.0', '1.0.0', 'use SaveLayoutAction')
    def __init__(self, *args, **kwargs):
        savelayout.SaveLayoutAction.__init__(self, *args, **kwargs)



#!/usr/bin/env python
#
# atlaspanel.py - The AtlasPanel class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`AtlasPanel`, a *FSLeyes control* panel
which allows the user to browse the FSL atlas images. See the
:mod:`~fsleyes` package documentation for more details on control panels,
and the :mod:`.atlases` module for more details on the atlases available in
FSL.
"""


import          logging

import numpy as np
import          wx

import fsl.data.image                as fslimage
import fsl.data.atlases              as atlases
import fsl.data.constants            as constants
import fsl.utils.idle                as idle

import fsleyes_props                 as props
import fsleyes_widgets.notebook      as notebook
import fsleyes_widgets.utils.status  as status

import fsleyes.controls.controlpanel as ctrlpanel
import fsleyes.strings               as strings
from . import                           atlasmanagementpanel
from . import                           atlasoverlaypanel
from . import                           atlasinfopanel


log = logging.getLogger(__name__)


class AtlasPanel(ctrlpanel.ControlPanel):
    """An ``AtlasPanel`` is a :class:`.ControlPanel` which allows the user to
    view atlas information, and to browse through the atlases that come shipped
    with FSL. The ``AtlasPanel`` interface is provided by some sub-panels,
    which are displayed in a :class:`fsleyes_widgets.Notebook` panel. The
    ``AtlasPanel`` itself provides a number of convenience methods that are
    used by these sub-panels:


    ============================== ===========================================
    :class:`.AtlasInfoPanel`       Displays information for the current
                                   :attr:`.DisplayContext.location` from
                                   atlases selected by the user.

    :class:`.AtlasOverlayPanel`    Allows the user to search through all
                                   atlases for specific regions, and to toggle
                                   on/off overlays for those regions.

    :class:`.AtlasManagementPanel` Allows the user to add/remove atlases.
    ============================== ===========================================


    **Loading atlases**


    The :class:`AtlasPanel` class provides the :meth:`loadAtlas` method, which
    is used by sub-panels to load atlas images.


    .. _atlas-panel-atlas-overlays:

    **Toggling atlas overlays**


    Both of the sub-panels allow the user to add/remove overlays to/from the
    :class:`.OverlayList`. The following overlay types can be added:

      - A complete summary :class:`.LabelAtlas`, which is a 3D image where
        each region has a discrete integer label. These images are added with
        a :attr:`.Display.overlayType` of ``label``.

      - A mask image containing a single region, extracted from a
        :class:`.LabelAtlas`. These images are added with a
        :attr:`.Display.overlayType` of ``mask``.

      - A 3D image containing the statistic image for a single region,
        extracted from a :class:`.StatisticAtlas`. These images are added
        with a :attr:`.Display.overlayType` of ``volume``.


    The following methods allow these overlays to be toggled on/off, and to
    query their state:

    .. autosummary::
       :nosignatures:

       toggleOverlay
       getOverlayName
       getOverlayState


    .. _atlas-panel-overlay-names:

    **Atlas overlay names**


    When an atlas overlay is added, its :attr:`.Image.name` (and subsequently
    its :attr:`.Display.name`) are set to a name which has the following
    structure::

        atlasID/overlayType/regionName

    where:

      - ``atlasID`` is the atlas identifier (see the :mod:`.atlases` module).

      - ``overlayType`` is either ``label``, ``prob``, or ``stat``, depending on
        whether the overlay is a discrete label image, a probaility image, or
        a statistic image..

      - ``regionName`` is the name of the region, or ``all`` if the overlay
        is a complete :class:`.LabelAtlas`.


    .. image:: images/atlaspanel_overlay_names.png
       :scale: 50%
       :align: center



    This name is used by the ``AtlasPanel`` to identify the overlay in the
    :class:`.OverlayList`.


    .. warning:: If the name of these overlays is changed, the ``AtlasPanel``
                 will not be able to find them in the :class:`.OverlayList`,
                 and the :meth:`toggleOverlay` and :meth:`getOverlayState`
                 methods will stop working properly.  So don't change the
                 atlas overlay names!


    **Locating regions**


    Finally, the :meth:`locateRegion` method allows the
    :attr:`.DisplayContext.location` to be moved to the location of a specific
    region in a specific atlas.
    """


    def __init__(self, parent, overlayList, displayCtx, frame):
        """Create an ``AtlasPanel``.

        :arg parent:      The :mod:`wx` parent object.
        :arg overlayList: The :class:`.OverlayList` instance.
        :arg displayCtx:  The :class:`.DisplayContext` instance.
        :arg frame:       The :class:`.FSLeyesFrame` instance.
        """

        ctrlpanel.ControlPanel.__init__(
            self, parent, overlayList, displayCtx, frame)

        # Make sure the atlas
        # registry is up to date
        atlases.rescanAtlases()

        # See the enableAtlasPanel method
        # for info about this attribute.
        self.__atlasPanelEnableStack = 0

        # Cache of loaded atlases
        # and enabled atlas overlays.
        self.__enabledOverlays = {}
        self.__loadedAtlases   = {}

        self.__notebook = notebook.Notebook(self)

        self.__sizer = wx.BoxSizer(wx.HORIZONTAL)
        self.__sizer.Add(self.__notebook, flag=wx.EXPAND, proportion=1)

        self.SetSizer(self.__sizer)

        self.__infoPanel = atlasinfopanel.AtlasInfoPanel(
            self.__notebook, overlayList, displayCtx, frame, self)

        # Overlay panel, containing a list of regions,
        # allowing the user to add/remove overlays
        self.__overlayPanel = atlasoverlaypanel.AtlasOverlayPanel(
            self.__notebook, overlayList, displayCtx, frame, self)

        self.__managePanel = atlasmanagementpanel.AtlasManagementPanel(
            self.__notebook, overlayList, displayCtx, frame, self)

        self.__notebook.AddPage(self.__infoPanel,
                                strings.titles[self.__infoPanel])
        self.__notebook.AddPage(self.__overlayPanel,
                                strings.titles[self.__overlayPanel])
        self.__notebook.AddPage(self.__managePanel,
                                strings.titles[self.__managePanel])

        self.overlayList.addListener('overlays',
                                     self.name,
                                     self.__overlayListChanged)

        self.Layout()
        self.SetMinSize(self.__sizer.GetMinSize())


    def destroy(self):
        """Must be called on when this ``AtlasPanel`` is no longer needed.

        Calls the ``destroy`` methods of the :class:`.AtlasInfoPanel` and
        :class:`.AtlasOverlayPanel`, and then calls
        :meth:`.ControlPanel.destroy`.
        """
        self.__loadedAtlases   = None
        self.__enabledOverlays = None
        self.__infoPanel     .destroy()
        self.__overlayPanel  .destroy()
        self.__managePanel   .destroy()

        self.overlayList.removeListener('overlays', self.name)

        ctrlpanel.ControlPanel.destroy(self)


    def Enable(self, enable=True):
        """Enables/disables this ``AtlasPanel``. """

        self.__infoPanel   .Enable(enable)
        self.__overlayPanel.Enable(enable)
        self.__managePanel .Enable(enable)


    def Disable(self):
        """Disables this ``AtlasPanel``. """
        self.Enable(False)


    def enableAtlasPanel(self, enable=True):
        """Disables/enables the :class:`.AtlasPanel` which contains this
        ``AtlasOverlayPanel``. This method is used by
        :class:`OverlayListWidget` instances.

        This method keeps a count of the number of times that it has been
        called - the count is increased every time a request is made
        to disable the ``AtlasPanel``, and decreased on requests to
        enable it. The ``AtlasPanel`` is only enabled when the count
        reaches 0.

        This ugly method solves an awkward problem - the ``AtlasOverlayPanel``
        disables the ``AtlasPanel`` when an atlas overlay is toggled on/off
        (via an ``OverlayListWidget``), and when an atlas region list is being
        generated (via the :meth:`__onAtlasSelect` method). If both of these
        things occur at the same time, the ``AtlasPanel`` could be prematurely
        re-enabled. This method overcomes this problem.
        """
        count = self.__atlasPanelEnableStack

        log.debug('enableAtlasPanel({}, count={})'.format(enable, count))

        if enable:
            count -= 1

            if count <= 0:
                count = 0
                self.Enable()

        else:
            count += 1
            self.Disable()

        self.__atlasPanelEnableStack = count


    def loadAtlas(self,
                  atlasID,
                  summary,
                  onLoad=None,
                  onError=None,
                  matchResolution=True):
        """Loads the atlas image with the specified ID. The atlas is loaded
        asynchronously (via the :mod:`.idle` module), as it can take some
        time. Use the `onLoad` argument if you need to do something when the
        atlas has been loaded.

        :arg onLoad:          Optional. A function which is called when the
                              atlas has been loaded, and which is passed the
                              loaded :class:`.Atlas` image.

        :arg onError:         Optional. A function which is called if the
                              atlas loading job raises an error. Passed the
                              ``Exception`` that was raised.

        :arg matchResolution: If ``True`` (the default), the version of the
                              atlas with the most suitable resolution, based
                              on the current contents of the
                              :class:`.OverlayList`, is loaded.

        See the :func:`.atlases.loadAtlas` function for details on the other
        arguments.
        """

        # Get the atlas description, and the
        # most suitable resolution to load.
        desc = atlases.getAtlasDescription(atlasID)
        res  = self.__getSuitableResolution(desc, matchResolution)

        if desc.atlasType == 'label':
            summary = True

        atlas = self.__loadedAtlases.get((atlasID, summary, res), None)

        if atlas is None:

            log.debug('Loading atlas {}/{}'.format(
                atlasID, 'label' if summary else 'prob'))

            status.update('Loading atlas {}...'.format(atlasID), timeout=None)

            def load():

                # the panel might get destroyed
                # before this function is called
                if self.destroyed:
                    return

                atlas = atlases.loadAtlas(atlasID, summary, resolution=res)

                # The atlas panel may be destroyed
                # before the atlas is loaded.
                if not self or self.destroyed:
                    return

                self.__loadedAtlases[atlasID, summary, res] = atlas

                status.update('Atlas {} loaded.'.format(atlasID))

                if onLoad is not None:
                    idle.idle(onLoad, atlas)

            idle.run(load, onError=onError)

        # If the atlas has already been loaded,
        # pass it straight to the onload function
        elif onLoad is not None:
            onLoad(atlas)


    def __getSuitableResolution(self, desc, matchResolution=True):
        """Used by the :meth:`loadAtlas` method. Determines a suitable
        atlas resolution to load, based on the current contents of the
        :class:`.OverlayList`.
        """

        niftis = [o for o in self.overlayList
                  if (isinstance(o, fslimage.Nifti) and
                      o.getXFormCode() == constants.NIFTI_XFORM_MNI_152)]

        # No overlays to match resolution against
        if len(niftis) == 0:
            matchResolution = False

        # If we don't need to match resolution,
        # return the highest available resolution
        # (the lowest value).
        if not matchResolution:
            return np.concatenate(desc.pixdims).min()

        # Find the highest resolution
        # in the overlay list
        pixdims = [o.pixdim[:3] for o in niftis]
        res     = np.concatenate(pixdims).min()

        # identify the atlas with the
        # nearest resolution to the
        # requested resolution
        reses = np.concatenate(desc.pixdims)
        res   = reses[np.argmin(np.abs(reses - res))]

        return res


    def getOverlayName(self, atlasID, labelIdx, summary):
        """Returns a name to be used for the specified atlas (see the section
        on :ref:`atlas names <atlas-panel-overlay-names>`).

        :arg atlasID:  Atlas identifier

        :arg labelIdx: Label index, or ``None`` for a complete atlas.

        :arg summary:  ``True`` corresponds to a label atlas, ``False`` to a
                       probabilistic atlas.
        """
        atlasDesc = atlases.getAtlasDescription(atlasID)

        if atlasDesc.atlasType == 'summary' or labelIdx is None:
            summary = True

        if summary: overlayType = 'label'
        else:       overlayType = 'prob'

        if labelIdx is None:
            overlayName = '{}/{}/all'.format(atlasID, overlayType)
        else:
            overlayName = '{}/{}/{}' .format(atlasID,
                                             overlayType,
                                             atlasDesc.labels[labelIdx].name)

        return overlayName, summary


    def getOverlayState(self, atlasID, labelIdx, summary):
        """Returns ``True`` if the specified atlas overlay is in the
        :class:`.OverlayList`, ``False`` otherwise.  See
        :meth:`getOverlayName` for details on the arguments.
        """

        name, _ = self.getOverlayName(atlasID, labelIdx, summary)
        return self.overlayList.find(name) is not None


    def toggleOverlay(self,
                      atlasID,
                      labelIdx,
                      summary,
                      onLoad=None,
                      onError=None):
        """Adds or removes the specified overlay to/from the
        :class:`.OverlayList`.

        :arg onLoad:  Optional function to be called when the overlay has been
                      added/removed.

        :arg onError: Optional function to be called if an error occurs while
                      loading an overlay.

        See :meth:`getOverlayName` for details on the other arguments.
        """

        atlasDesc            = atlases.getAtlasDescription(atlasID)
        overlayName, summary = self.getOverlayName(atlasID, labelIdx, summary)
        overlay              = self.overlayList.find(overlayName)

        if overlay is not None:

            self.overlayList.disableListener('overlays', self.name)
            self.overlayList.remove(overlay)
            self.overlayList.enableListener('overlays', self.name)

            self.__enabledOverlays.pop(overlayName, None)
            self.__overlayPanel.setOverlayState(
                atlasDesc, labelIdx, summary, False)

            log.debug('Removed overlay {}'.format(overlayName))

            if onLoad is not None:
                onLoad()
            return

        def realOnLoad(atlas):
            initprops = {}
            # label image
            if labelIdx is None:
                overlay = fslimage.Image(atlas)
                initprops['overlayType'] = 'label'

            else:

                # regional label image
                if summary:
                    overlay = atlas.get(index=labelIdx, binary=False)
                    initprops['overlayType'] = 'mask'
                    initprops['colour']      = np.random.random(3)

                # regional statistic/probability image
                else:
                    overlay = atlas.get(index=labelIdx)
                    initprops['overlayType']   = 'volume'
                    initprops['cmap']          = 'hot'
                    initprops['displayRange']  = (atlasDesc.lower,
                                                  atlasDesc.upper)
                    initprops['clippingRange'] = (atlasDesc.lower,
                                                  atlasDesc.upper)

            overlay.name = overlayName

            with props.suppress(self.overlayList, 'overlays', self.name):
                self.overlayList.append(overlay, **initprops)

            self.__overlayPanel.setOverlayState(
                atlasDesc, labelIdx, summary, True)

            self.__enabledOverlays[overlayName] = (overlay,
                                                   atlasID,
                                                   labelIdx,
                                                   summary)

            log.debug('Added overlay {}'.format(overlayName))

            if onLoad is not None:
                onLoad()

        self.loadAtlas(atlasID, summary, onLoad=realOnLoad, onError=onError)


    def locateRegion(self, atlasID, labelIdx):
        """Moves the :attr:`.DisplayContext.location`  to the specified
        region in the specified atlas. See the :class:`.AtlasDescription`
        class for details on atlas identifiers/label indices.

        :arg atlasID:  Atlas identifier
        :arg labelIdx: Label index
        """

        atlasDesc = atlases.getAtlasDescription(atlasID)
        label     = atlasDesc.labels[labelIdx]
        overlay   = self.displayCtx.getReferenceImage(
            self.displayCtx.getSelectedOverlay())

        if overlay is None:
            log.warn('No reference image available - cannot locate region')

        opts     = self.displayCtx.getOpts(overlay)
        worldLoc = (label.x, label.y, label.z)
        dispLoc  = opts.transformCoords([worldLoc], 'world', 'display')[0]

        self.displayCtx.location.xyz = dispLoc


    def __overlayListChanged(self, *a):
        """Called when the :class:`.OverlayList` changes.

        Makes sure that the :class:`.AtlasOverlayPanel` state is up to date -
        see the :meth:`.AtlasOverlayPanel.setOverlayState` method.
        """

        for overlayName in list(self.__enabledOverlays.keys()):

            overlay, atlasID, labelIdx, summary = \
                self.__enabledOverlays[overlayName]

            if overlay not in self.overlayList:

                self.__enabledOverlays.pop(overlayName)
                atlasDesc = atlases.getAtlasDescription(atlasID)
                self.__overlayPanel.setOverlayState(
                    atlasDesc, labelIdx, summary, False)



#!/usr/bin/env python
#
# profilemap.py - CanvasPanel -> Profile mappings.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module is used by the :class:`.Profile` and :class:`.ProfileManager`
classes.

It defines a few dictionaries which define the profile type to use for each
:class:`.ViewPanel` type, temporary mouse/keyboard interaction modes, and
alternate mode handlers for the profiles contained in the profiles package.

.. autosummary::
   profiles
   profileHandlers
   tempModeMap
   altHandlerMap
"""

import logging

from collections import OrderedDict

import wx

from fsleyes.views.orthopanel             import OrthoPanel
from fsleyes.views.lightboxpanel          import LightBoxPanel
from fsleyes.views.timeseriespanel        import TimeSeriesPanel
from fsleyes.views.histogrampanel         import HistogramPanel
from fsleyes.views.powerspectrumpanel     import PowerSpectrumPanel
from fsleyes.views.scene3dpanel           import Scene3DPanel

from fsleyes.profiles.orthoviewprofile    import OrthoViewProfile
from fsleyes.profiles.orthoeditprofile    import OrthoEditProfile
from fsleyes.profiles.orthocropprofile    import OrthoCropProfile
from fsleyes.profiles.lightboxviewprofile import LightBoxViewProfile
from fsleyes.profiles.plotprofile         import PlotProfile
from fsleyes.profiles.histogramprofile    import HistogramProfile
from fsleyes.profiles.timeseriesprofile   import TimeSeriesProfile
from fsleyes.profiles.scene3dviewprofile  import Scene3DViewProfile


log = logging.getLogger(__name__)


profiles  = {
    OrthoPanel         : ['view', 'edit', 'crop'],
    LightBoxPanel      : ['view'],
    TimeSeriesPanel    : ['view'],
    HistogramPanel     : ['view'],
    PowerSpectrumPanel : ['view'],
    Scene3DPanel       : ['view'],
}
"""This dictionary is used by the :class:`.ProfileManager` to figure out which
profiles are available for each :class:`.ViewPanel`. They are added as options
to the :attr:`.ViewPanel.profile` property.
"""


profileHandlers = {
    (OrthoPanel,         'view') : OrthoViewProfile,
    (OrthoPanel,         'edit') : OrthoEditProfile,
    (OrthoPanel,         'crop') : OrthoCropProfile,
    (LightBoxPanel,      'view') : LightBoxViewProfile,
    (TimeSeriesPanel,    'view') : TimeSeriesProfile,
    (HistogramPanel,     'view') : HistogramProfile,
    (PowerSpectrumPanel, 'view') : PlotProfile,
    (Scene3DPanel,       'view') : Scene3DViewProfile,
}
"""This dictionary is used by the :class:`.ProfileManager` class to figure out
which :class:`.Profile` sub-class to create for a given :class:`.ViewPanel`
instance and profile identifier.
"""


# Important: Any temporary modes which use CTRL,
# ALT, or CTRL+ALT must not handle character events,
# as these modifiers are reserved for global
# shortcuts.


# For multi-key combinations, the modifier key
# IDs must be provided as a tuple, in alphabetical
# order. For example, to specify shift+ctrl, the
# tuple must be (wx.WXK_CTRL, wx.WXK_SHIFT)
tempModeMap = {

    # Command/CTRL puts the user in zoom mode,
    # and ALT puts the user in pan mode
    OrthoViewProfile : OrderedDict((
        (('nav',   wx.WXK_CONTROL),                'zoom'),
        (('nav',   wx.WXK_ALT),                    'pan'),
        (('nav',   wx.WXK_SHIFT),                  'slice'),
        (('nav',  (wx.WXK_CONTROL, wx.WXK_SHIFT)), 'bricon'))),

    # OrthoEditProfile inherits all of the
    # settings for OrthoViewProfile above,
    # but overrides a few key ones.
    OrthoEditProfile : OrderedDict((

        (('sel',     wx.WXK_SHIFT),                  'slice'),
        (('desel',   wx.WXK_SHIFT),                  'slice'),
        (('selint',  wx.WXK_SHIFT),                  'slice'),
        (('fill',    wx.WXK_SHIFT),                  'slice'),
        (('sel',     wx.WXK_ALT),                    'pan'),
        (('desel',   wx.WXK_ALT),                    'pan'),
        (('selint',  wx.WXK_ALT),                    'pan'),
        (('fill',    wx.WXK_ALT),                    'pan'),
        (('sel',     wx.WXK_CONTROL),                'zoom'),
        (('desel',   wx.WXK_CONTROL),                'zoom'),
        (('selint',  wx.WXK_CONTROL),                'zoom'),
        (('fill',    wx.WXK_CONTROL),                'zoom'),
        (('sel',    (wx.WXK_CONTROL, wx.WXK_SHIFT)), 'chsize'),
        (('desel',  (wx.WXK_CONTROL, wx.WXK_SHIFT)), 'chsize'),
        (('selint', (wx.WXK_CONTROL, wx.WXK_SHIFT)), 'chthres'),
        (('selint', (wx.WXK_ALT,     wx.WXK_SHIFT)), 'chrad'))),

    OrthoCropProfile : OrderedDict((

        (('crop',  wx.WXK_SHIFT),                  'nav'),
        (('crop',  wx.WXK_CONTROL),                'zoom'),
        (('crop',  wx.WXK_ALT),                    'pan'),
        (('crop', (wx.WXK_CONTROL, wx.WXK_SHIFT)), 'slice'),
    )),

    LightBoxViewProfile : OrderedDict((
        (('view', wx.WXK_CONTROL), 'zoom'), )),

    # Can't use shift on mpl
    # canvases for some reason.
    TimeSeriesProfile : OrderedDict((
        (('panzoom', wx.WXK_CONTROL), 'volume'),
    )),

    HistogramProfile : OrderedDict((
        (('panzoom', wx.WXK_CONTROL), 'overlayRange'),
    )),

    Scene3DViewProfile : OrderedDict((
        (('rotate', wx.WXK_CONTROL), 'zoom'),
        (('rotate', wx.WXK_ALT),     'pan'),
        (('rotate', wx.WXK_SHIFT),   'pick'),
    )),
}
"""The ``tempModeMap`` dictionary defines temporary modes, for each
:class:`Profile` sub-class which, when in a given mode, can be accessed with a
keyboard modifer (e.g. Control, Shift, etc). For example, a temporary mode map
of::

    ('view', wx.WXK_SHIFT) : 'zoom'

states that when the ``Profile`` is in ``'view'`` mode, and the shift key is
held down, the ``Profile`` should temporarily switch to ``'zoom'`` mode.
"""


altHandlerMap = {

    OrthoViewProfile : OrderedDict((

        # in navigate, slice, and zoom mode, the
        # left mouse button navigates, the right
        # mouse button draws a zoom rectangle,
        # and the middle button pans.
        (('nav',  'LeftMouseDown'),   ('nav',  'LeftMouseDrag')),
        (('nav',  'MiddleMouseDrag'), ('pan',  'LeftMouseDrag')),
        (('nav',  'RightMouseDown'),  ('zoom', 'RightMouseDown')),
        (('nav',  'RightMouseDrag'),  ('zoom', 'RightMouseDrag')),
        (('nav',  'RightMouseUp'),    ('zoom', 'RightMouseUp')),

        # In slice mode, the left and right
        # mouse buttons work as for pick mode.
        # Middle and right mouse works the
        # same as for nav mode, defined above.
        (('slice', 'LeftMouseDown'),   ('pick', 'LeftMouseDown')),
        (('slice', 'LeftMouseDrag'),   ('pick', 'LeftMouseDrag')),
        (('slice', 'MiddleMouseDrag'), ('pan',  'LeftMouseDrag')),
        (('slice', 'RightMouseDown'),  ('zoom', 'RightMouseDown')),
        (('slice', 'RightMouseDrag'),  ('zoom', 'RightMouseDrag')),
        (('slice', 'RightMouseUp'),    ('zoom', 'RightMouseUp')),

        # In zoom mode, the left mouse button
        # navigates, the right mouse button
        # draws a zoom rectangle, and the
        # middle mouse button pans
        (('zoom', 'RightMouseDown'),  ('zoom', 'RightMouseDrag')),
        (('zoom', 'LeftMouseDown'),   ('nav',  'LeftMouseDown')),
        (('zoom', 'LeftMouseDrag'),   ('nav',  'LeftMouseDrag')),
        (('zoom', 'MiddleMouseDrag'), ('pan',  'LeftMouseDrag')),

        # In pick mode, left mouse down
        # is the same as left mouse drag.
        (('pick', 'LeftMouseDown'),   ('pick', 'LeftMouseDrag')))),

    OrthoEditProfile : OrderedDict((

        # The OrthoEditProfile is in
        # 'nav' mode by default.

        # View keyboard navigation
        # works in major edit modes
        (('sel',    'Char'),  ('nav', 'Char')),
        (('desel',  'Char'),  ('nav', 'Char')),
        (('selint', 'Char'),  ('nav', 'Char')),
        (('fill',   'Char'),  ('nav', 'Char')),

        # When in select mode, the right
        # mouse button allows the user
        # to deselect voxels.
        (('sel',    'RightMouseDown'),  ('desel',  'LeftMouseDown')),
        (('sel',    'RightMouseDrag'),  ('desel',  'LeftMouseDrag')),
        (('sel',    'RightMouseUp'),    ('desel',  'LeftMouseUp')),

        # And vice versa
        (('desel',  'RightMouseDown'),  ('sel',    'LeftMouseDown')),
        (('desel',  'RightMouseDrag'),  ('sel',    'LeftMouseDrag')),
        (('desel',  'RightMouseUp'),    ('sel',    'LeftMouseUp')),

        # Same for fill mode
        (('fill',   'RightMouseDown'),  ('desel',  'LeftMouseDown')),
        (('fill',   'RightMouseDrag'),  ('desel',  'LeftMouseDrag')),
        (('fill',   'RightMouseUp'),    ('desel',  'LeftMouseUp')),

        # TODO Need a way to navigate?
        (('selint', 'RightMouseDown'),  ('desel',  'LeftMouseDown')),
        (('selint', 'RightMouseDrag'),  ('desel',  'LeftMouseDrag')),
        (('selint', 'RightMouseUp'),    ('desel',  'LeftMouseUp')),

        # Make the selection cursor
        # visible in desel mode
        (('desel',  'MouseMove'),       ('sel',    'MouseMove')),

        # Middle mouse always pans.
        (('sel',    'MiddleMouseDrag'), ('pan', 'LeftMouseDrag')),
        (('desel',  'MiddleMouseDrag'), ('pan', 'LeftMouseDrag')),
        (('selint', 'MiddleMouseDrag'), ('pan', 'LeftMouseDrag')),
    )),

    OrthoCropProfile : OrderedDict((
        (('crop', 'MiddleMouseDrag'), ('pan', 'LeftMouseDrag')), )),

    LightBoxViewProfile : OrderedDict((
        (('view', 'LeftMouseDown'), ('view', 'LeftMouseDrag')), )),

    Scene3DViewProfile : OrderedDict((
        (('rotate', 'MiddleMouseDown'), ('pan', 'LeftMouseDown')),
        (('rotate', 'MiddleMouseDrag'), ('pan', 'LeftMouseDrag')),
        (('rotate', 'MiddleMouseUp'),   ('pan', 'LeftMouseUp')),
    )),


    # We cannot remap mouse buttons on the
    # PlotProfile.panzoom mode, because the
    # mpl NavigationToolbar2 class has
    # hard-coded left/right mouse button
    # behaviours.
}
"""The ``altHandlerMap`` dictionary defines alternate handlers for a given
mode and event type. Entries in this dictionary allow a :class:`.Profile`
sub-class to define a handler for a single mode and event type, but to re-use
that handler for other modes and event types. For example, the following
alternate handler mapping::

    ('zoom', 'MiddleMouseDrag'), ('pan',  'LeftMouseDrag'))

states that when the ``Profile`` is in ``'zoom'`` mode, and a
``MiddleMouseDrag`` event occurs, the ``LeftMouseDrag`` handler for the
``'pan'`` mode should be called.

.. note:: Event bindings defined in the ``altHandlerMap`` take precdence over
          the event bindings defined in the :class:`.Profile` sub-class. So
          you can use the ``altHandlerMap`` to override the default behaviour
          of a ``Profile``.
"""


fallbackHandlerMap = {
    OrthoViewProfile : OrderedDict((
        (('pick', 'LeftMouseDown'), ('nav', 'LeftMouseDown')),
        (('pick', 'LeftMouseDrag'), ('nav', 'LeftMouseDrag')),
    ))
}
"""The ``fallbackHandlerMap`` dictionary defines handlers for a given mode and
event type which will be called if the handler for that mode/event type returns
a value of ``False``, indicating that it has not been handled.
"""



#!/usr/bin/env python
#
# splash.py - FSLeyes splash screen.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`FSLeyesSplash` class, a splash screen for
*FSLeyes*.
"""


import            time
import os.path as op

import wx

import                               fsleyes
import fsleyes_widgets.imagepanel as imagepanel
import fsleyes.strings            as strings


def getSplashFile():
    """Returns the path to the splash screen image file. """
    return op.join(fsleyes.assetDir, 'assets', 'icons', 'splash', 'splash.png')


class FSLeyesSplash(wx.Frame):
    """A simple splash screen for *FSLeyes*. An image and a status bar are
    displayed; the status bar can be updated via the :meth:`SetStatus` method.


    The :class:`.ImagePanel` class is used to display the image.


    Typical usage would be something like the following::

        splash = FSLeyesSplash(None)
        splash.Show()

        # Do something, e.g. loading overlays
        splash.SetStatus('Loading blah.nii.gz ...')

        # Finished initialising, the application is ready
        splash.Close()
    """

    def __init__(self, parent):
        """Create a ``FSLeyesSplash`` frame.

        :arg parent: The :mod:`wx` parent object.
        """

        wx.Frame.__init__(self, parent, style=wx.FULL_REPAINT_ON_RESIZE)

        splashbmp  = wx.Bitmap(getSplashFile(), wx.BITMAP_TYPE_PNG)
        splashimg  = splashbmp.ConvertToImage()

        self.__splashPanel = imagepanel.ImagePanel(self, splashimg)
        self.__splashPanel.SetMinSize(splashimg.GetSize())
        self.__statusBar   = wx.StaticText(self, style=wx.ST_ELLIPSIZE_MIDDLE)

        self.__statusBar.SetLabel(strings.messages[self, 'default'])

        self.__statusBar.SetBackgroundColour('#000000')
        self.__statusBar.SetForegroundColour('#ffffff')
        self            .SetBackgroundColour('#000000')

        self.__sizer = wx.BoxSizer(wx.VERTICAL)

        self.__sizer.Add(self.__splashPanel, flag=wx.EXPAND, proportion=1)
        self.__sizer.Add(self.__statusBar,   flag=wx.EXPAND)

        self.SetSizer(self.__sizer)

        self.Layout()
        self.Fit()


    def Show(self):
        """Show this ``FSLeyesSplash`` frame, and centre it on the screen. """

        wx.Frame.Show(self)
        self.CentreOnScreen()
        self.Refresh()
        self.Update()

        # GTK is a piece of shit. Refresh/Update, combined
        # with a straight call to Yield does not guarantee
        # that the splash screen will be displayed. It
        # seems that a short delay is necessary before it
        # will be drawn, during which time we can't do
        # anything which would block the application loop.
        for i in range(10):
            wx.GetApp().Yield()
            time.sleep(0.025)


    def SetStatus(self, text):
        """Sets the text shown on the status bar to the specified ``text``. """
        self.__statusBar.SetLabel(text)
        self.__statusBar.Refresh()
        self.__statusBar.Update()



#!/usr/bin/env python
#
# glmesh_funcs.py - OpenGL 2.1 functions used by the GLMesh class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides functions which are used by the :class:`.GLMesh`
class to render :class:`.Mesh` overlays in an OpenGL 2.1 compatible
manner.

A :class:`.GLSLShader` is used to manage the ``glmesh`` vertex/fragment
shader programs.
"""


import OpenGL.GL as gl

import fsleyes.gl.shaders as shaders


def compileShaders(self):
    """Loads the ``glmesh`` vertex/fragment shader source and creates
    :class:`.GLSLShader` instance(s).
    """

    if self.threedee:

        flatVertSrc = shaders.getVertexShader(  'glmesh_3d_flat')
        flatFragSrc = shaders.getFragmentShader('glmesh_3d_flat')
        dataVertSrc = shaders.getVertexShader(  'glmesh_3d_data')
        dataFragSrc = shaders.getFragmentShader('glmesh_3d_data')

        self.flatShader = shaders.GLSLShader(flatVertSrc,
                                             flatFragSrc,
                                             indexed=True)
        self.dataShader = shaders.GLSLShader(dataVertSrc,
                                             dataFragSrc,
                                             indexed=True)

    else:

        vertSrc = shaders.getVertexShader(  'glmesh_2d_data')
        fragSrc = shaders.getFragmentShader('glmesh_2d_data')

        self.dataShader = shaders.GLSLShader(vertSrc, fragSrc)


def updateShaderState(self, **kwargs):
    """Updates the shader program according to the current :class:`.MeshOpts``
    configuration.
    """

    dopts    = self.opts
    copts    = self.canvas.opts
    dshader  = self.dataShader
    fshader  = self.flatShader

    dshader.load()
    dshader.set('cmap',           0)
    dshader.set('negCmap',        1)
    dshader.set('useNegCmap',     kwargs['useNegCmap'])
    dshader.set('cmapXform',      kwargs['cmapXform'])
    dshader.set('flatColour',     kwargs['flatColour'])
    dshader.set('invertClip',     dopts.invertClipping)
    dshader.set('discardClipped', dopts.discardClipped)
    dshader.set('modulateAlpha',  dopts.modulateAlpha)
    dshader.set('modScale',       kwargs['modScale'])
    dshader.set('modOffset',      kwargs['modOffset'])
    dshader.set('clipLow',        dopts.clippingRange.xlo)
    dshader.set('clipHigh',       dopts.clippingRange.xhi)

    if self.threedee:
        dshader.set('lighting', copts.light)
        dshader.set('lightPos', kwargs['lightPos'])

        dshader.setAtt('vertex', self.vertices)
        dshader.setAtt('normal', self.normals)

        vdata = self.getVertexData('vertex')
        mdata = self.getVertexData('modulate')

        # if modulate data is not set,
        # we use the vertex data
        if mdata is None:
            mdata = vdata

        if vdata is not None: dshader.setAtt('vertexData',   vdata.ravel('C'))
        if mdata is not None: dshader.setAtt('modulateData', mdata.ravel('C'))

        dshader.setIndices(self.indices)

    dshader.unload()

    if self.threedee:
        fshader.load()
        fshader.set('lighting', copts.light)
        fshader.set('lightPos', kwargs['lightPos'])
        fshader.set('colour',   kwargs['flatColour'])

        fshader.setAtt('vertex', self.vertices)
        fshader.setAtt('normal', self.normals)
        fshader.setIndices(self.indices)
        fshader.unload()


def preDraw(self):
    """Must be called before :func:`draw`. Loads the appropriate shader
    program.
    """

    flat = self.opts.vertexData is None

    if flat: shader = self.flatShader
    else:    shader = self.dataShader

    self.activeShader = shader
    shader.load()


def draw(self,
         glType,
         vertices,
         indices=None,
         normals=None,
         vdata=None,
         mdata=None):
    """Called for 3D meshes, and when :attr:`.MeshOpts.vertexData` is not
    ``None``. Loads and runs the shader program.

    :arg glType:   The OpenGL primitive type.

    :arg vertices: ``(n, 3)`` array containing the line vertices to draw.

    :arg indices:  Indices into the ``vertices`` array. If not provided,
                   ``glDrawArrays`` is used.

    :arg normals:  Vertex normals.

    :arg vdata:    ``(n, )`` array containing data for each vertex.

    :arg mdata:    ``(n, )`` array containing alpha modulation data for
                   each vertex.
    """

    shader = self.activeShader

    # for 3D, shader attributes are
    # configured in updateShaderState
    if self.threedee:
        vertices = None
        normals  = None
        vdata    = None
        mdata    = None

    if vertices is not None: shader.setAtt('vertex',       vertices)
    if normals  is not None: shader.setAtt('normal',       normals)
    if vdata    is not None: shader.setAtt('vertexData',   vdata)
    if mdata    is not None: shader.setAtt('modulateData', mdata)

    shader.loadAtts()

    if indices is None:
        gl.glDrawArrays(glType, 0, vertices.shape[0])
    else:
        nverts = indices.shape[0]
        if self.threedee:
            indices = None
        gl.glDrawElements(glType, nverts, gl.GL_UNSIGNED_INT, indices)


def postDraw(self):
    """Must be called after :func:`draw`. Unloads shaders, and unbinds
    textures.
    """

    shader = self.activeShader
    shader.unloadAtts()
    shader.unload()
    self.activeShader = None



#!/usr/bin/env python
#
# shortcuts.py - Keyboard shortcuts
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module defines keyboard shortcuts used throughout FSLeyes.
"""


import fsleyes_widgets.utils.typedict as td


actions = td.TypeDict({

    'LoadOverlayAction'        : 'Ctrl-O',
    'LoadOverlayFromDirAction' : 'Ctrl-D',
    'LoadStandardAction'       : 'Ctrl-S',
    'NewImageAction'           : 'Ctrl-Shift-N',
    'CopyOverlayAction'        : 'Ctrl-Shift-C',
    'SaveOverlayAction'        : 'Ctrl-Shift-S',
    'ReloadOverlayAction'      : 'Ctrl-Shift-R',
    'RemoveOverlayAction'      : 'Ctrl-Shift-W',

    'FSLeyesFrame.closeFSLeyes'           : 'Ctrl-Q',
    'FSLeyesFrame.openHelp'               : 'Ctrl-?',

    'FSLeyesFrame.layouts.default'        : 'Ctrl-Shift-D',

    'FSLeyesFrame.addOrthoPanel'          : 'Ctrl-1',
    'FSLeyesFrame.addLightBoxPanel'       : 'Ctrl-2',
    'FSLeyesFrame.addTimeSeriesPanel'     : 'Ctrl-3',
    'FSLeyesFrame.addHistogramPanel'      : 'Ctrl-4',
    'FSLeyesFrame.addPowerSpectrumPanel'  : 'Ctrl-5',
    'FSLeyesFrame.addScene3DPanel'        : 'Ctrl-6',
    'FSLeyesFrame.addShellPanel'          : 'Ctrl-7',

    'FSLeyesFrame.selectNextOverlay'       : 'Ctrl-Up',
    'FSLeyesFrame.selectPreviousOverlay'   : 'Ctrl-Down',
    'FSLeyesFrame.toggleOverlayVisibility' : 'Ctrl-F',

    # Shortcuts for next/prev volume

    # ViewPanel actions must use one
    # of CTRL, ALT or Shift due to
    # hacky things in FSLeyesFrame
    # (see __onViewPanelMenuItem)

    'ViewPanel.removeFromFrame'             : 'Ctrl-W',
    'ViewPanel.removeAllPanels'             : 'Ctrl-Alt-X',

    'CanvasPanel.toggleOverlayList'         : 'Ctrl-Alt-1',
    'CanvasPanel.toggleLocationPanel'       : 'Ctrl-Alt-2',
    'CanvasPanel.toggleOverlayInfo'         : 'Ctrl-Alt-3',
    'CanvasPanel.toggleDisplayPanel'        : 'Ctrl-Alt-4',
    'CanvasPanel.toggleCanvasSettingsPanel' : 'Ctrl-Alt-5',
    'CanvasPanel.toggleAtlasPanel'          : 'Ctrl-Alt-6',

    'CanvasPanel.toggleDisplayToolBar'      : 'Ctrl-Alt-7',

    'OrthoPanel.toggleOrthoToolBar'         : 'Ctrl-Alt-8',
    'LightBoxPanel.toggleLightBoxToolBar'   : 'Ctrl-Alt-8',
    'Scene3DPanel.toggleScene3DToolBar'     : 'Ctrl-Alt-8',

    'CanvasPanel.toggleFileTreePanel'       : 'Ctrl-Alt-9',

    'CanvasPanel.toggleMovieMode'           : 'Alt-M',
    'CanvasPanel.toggleDisplaySync'         : 'Alt-S',

    'OrthoPanel.toggleEditMode'             : 'Alt-E',
    'OrthoPanel.resetDisplay'               : 'Alt-R',
    'OrthoPanel.centreCursor'               : 'Alt-P',
    'OrthoPanel.centreCursorWorld'          : 'Alt-O',
    'OrthoPanel.toggleLabels'               : 'Alt-L',
    'OrthoPanel.toggleCursor'               : 'Alt-C',
    'OrthoPanel.toggleXCanvas'              : 'Alt-X',
    'OrthoPanel.toggleYCanvas'              : 'Alt-Y',
    'OrthoPanel.toggleZCanvas'              : 'Alt-Z',
    'OrthoPanel.pearsonCorrelation'         : 'Alt-I',

    'PlotPanel.toggleOverlayList'         : 'Ctrl-Alt-1',
    'PlotPanel.togglePlotList'            : 'Ctrl-Alt-2',
    'PlotPanel.importDataSeries'          : 'Ctrl-I',
    'PlotPanel.exportDataSeries'          : 'Ctrl-E',

    'TimeSeriesPanel.toggleTimeSeriesToolBar'       : 'Ctrl-Alt-3',
    'TimeSeriesPanel.toggleTimeSeriesControl'       : 'Ctrl-Alt-4',
    'HistogramPanel.toggleHistogramToolBar'         : 'Ctrl-Alt-3',
    'HistogramPanel.toggleHistogramControl'         : 'Ctrl-Alt-4',
    'PowerSpectrumPanel.togglePowerSpectrumToolBar' : 'Ctrl-Alt-3',
    'PowerSpectrumPanel.togglePowerSpectrumControl' : 'Ctrl-Alt-4',

    'OrthoEditProfile.undo'           : 'Ctrl-Z',
    'OrthoEditProfile.redo'           : 'Ctrl-Y',
    'OrthoEditProfile.createMask'     : 'Ctrl-N',
    'OrthoEditProfile.clearSelection' : 'Ctrl-Shift-A',
    'OrthoEditProfile.eraseSelection' : 'Ctrl-E',
    'OrthoEditProfile.fillSelection'  : 'Ctrl-B',
    'OrthoEditProfile.copySelection'  : 'Ctrl-C',
    'OrthoEditProfile.pasteSelection' : 'Ctrl-V',
})
